from typing import Optional, Tuple, List, Dict, Any

from powl.discovery.total_order_based.inductive.variants.brute_force.factory import CutFactoryPOWLBruteForce
from powl.discovery.total_order_based.inductive.variants.im_tree import IMBasePOWL, T
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import POWLDiscoveryVariant
from powl.objects.obj import POWL


class POWLInductiveMinerBruteForce(IMBasePOWL):

    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.BRUTE_FORCE

    def find_cut(self, obj: T, parameters: Optional[Dict[str, Any]] = None) -> Optional[Tuple[POWL, List[T]]]:
        res = CutFactoryPOWLBruteForce.find_cut(obj, parameters=parameters)
        return res



from collections import Counter
from multiprocessing import Pool, Manager
from typing import Optional, Tuple, List, Dict, Any

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.algo.discovery.inductive.fall_through.strict_tau_loop import StrictTauLoopUVCL
from powl.objects.obj import OperatorPOWL
from pm4py.objects.process_tree.obj import Operator


class POWLStrictTauLoopUVCL(StrictTauLoopUVCL):

    @classmethod
    def apply(cls, obj: IMDataStructureUVCL, pool: Pool = None, manager: Manager = None, parameters: Optional[Dict[str, Any]] = None) -> Optional[Tuple[OperatorPOWL, List[IMDataStructureUVCL]]]:
        log = obj.data_structure
        proj = cls._get_projected_log(log)
        if sum(proj.values()) > sum(log.values()):
            return OperatorPOWL(Operator.LOOP, []), [IMDataStructureUVCL(proj), IMDataStructureUVCL(Counter())]




from abc import ABC
from collections import Counter
from itertools import combinations, product
from typing import Optional, List, Any, Generic, Dict, Collection, Tuple

from pm4py.algo.discovery.inductive.cuts.abc import Cut, T
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.objects.dfg import util as dfu
from pm4py.algo.discovery.inductive.cuts import utils as cut_util
from powl.objects.BinaryRelation import BinaryRelation
from powl.objects.obj import DecisionGraph, POWL
from pm4py.objects.process_tree.obj import Operator


class MaximalDecisionGraphCut(Cut[T], ABC, Generic[T]):
    @classmethod
    def operator(cls, parameters: Optional[Dict[str, Any]] = None) -> Operator:
        return None

    @classmethod
    def holds(cls, obj: T, parameters: Optional[Dict[str, Any]] = None) -> Optional[List[Any]]:

        alphabet = parameters["alphabet"]
        transitive_successors = parameters["transitive_successors"]

        groups = [frozenset([a]) for a in alphabet]

        for a, b in combinations(alphabet, 2):
            if b in transitive_successors[a] and a in transitive_successors[b]:
                groups = cut_util.merge_groups_based_on_activities(a, b, groups)

        if len(groups) < 2:
            return None

        # merged = True
        # while merged:
        #     merged = False
        #     new_groups = [g for g in groups]
        #     presets = {g: set() for g in groups}
        #     postsets = {g: set() for g in groups}
        #     for i, g1 in enumerate(groups):
        #         for j, g2 in enumerate(groups):
        #             if i != j:
        #                 pairs = product(g1, g2)
        #                 if any((a, b) in dfg.graph for (a, b) in pairs):
        #                     presets[g2] = presets[g2].union(g1)
        #                     postsets[g1] = postsets[g1].union(g2)
        #     for i, g1 in enumerate(groups):
        #         for j, g2 in enumerate(groups):
        #             if i != j:
        #                 if presets[g1] == presets[g2] and postsets[g1] == postsets[g2]:
        #                     new_groups = cut_util.merge_groups_based_on_activities(list(g1)[0], list(g2)[0], new_groups)
        #                     merged = True
        #     if len(new_groups) < 2:
        #         return groups
        #     else:
        #         groups = new_groups
        #
        # if len(groups) < 2:
        #     return None

        return groups

    @classmethod
    def apply(cls, obj: T, parameters: Optional[Dict[str, Any]] = None) -> Optional[Tuple[DecisionGraph,
    List[POWL]]]:

        dfg = obj.dfg
        alphabet = sorted(dfu.get_vertices(dfg), key=lambda g: g.__str__())

        start_activities = set(obj.dfg.start_activities.keys())
        end_activities = set(obj.dfg.end_activities.keys())

        transitive_predecessors, transitive_successors = dfu.get_transitive_relations(dfg)

        for a in alphabet:
            if a not in start_activities and len(set(transitive_predecessors[a]) & start_activities) == 0:
                alphabet = [elm for elm in alphabet if elm != a]
            if a not in end_activities and len(set(transitive_successors[a]) & end_activities) == 0:
                alphabet = [elm for elm in alphabet if elm != a]

        parameters["alphabet"] = alphabet
        parameters["transitive_successors"] = transitive_successors

        groups = cls.holds(obj, parameters)
        if groups is None:
            return groups
        children = cls.project(obj, groups, parameters)

        order = BinaryRelation(nodes=children)
        for i, g1 in enumerate(groups):
            for j, g2 in enumerate(groups):
                if i != j:
                    pairs = product(g1, g2)
                    if any((a, b) in dfg.graph for (a, b) in pairs):
                        order.add_edge(children[i], children[j])




        start_nodes = []
        end_nodes = []
        for i in range(len(groups)):
            node = groups[i]
            if any(a in start_activities for a in node):
                start_nodes.append(children[i])
            if any(a in end_activities for a in node):
                end_nodes.append(children[i])

        dg = DecisionGraph(order, start_nodes, end_nodes)
        return dg, dg.children


class MaximalDecisionGraphCutUVCL(MaximalDecisionGraphCut[IMDataStructureUVCL], ABC):
    @classmethod
    def project(cls, obj: IMDataStructureUVCL, groups: List[Collection[Any]],
                parameters: Optional[Dict[str, Any]] = None) -> List[
        IMDataStructureUVCL]:

        logs = [Counter() for _ in groups]

        for t in obj.data_structure:
            for i, group in enumerate(groups):
                proj = tuple(e for e in t if e in group)
                if proj:
                    logs[i].update({proj: obj.data_structure[t]})

        return [IMDataStructureUVCL(l) for l in logs]




from enum import Enum, auto


class POWLDiscoveryVariant(Enum):
    TREE = auto()  # base IM with no partial orders
    BRUTE_FORCE = auto()
    MAXIMAL = auto()
    DYNAMIC_CLUSTERING = auto()
    DECISION_GRAPH_MAX = auto()
    DECISION_GRAPH_CLUSTERING = auto()



from powl.discovery.partial_order_based.utils.combine_order import combine_orders
from powl.discovery.partial_order_based.utils.simplified_objects import Skip, Graph, LOOP, XOR, ActivityInstance, \
    generate_powl, SelfLoop, get_leaves
from powl.discovery.partial_order_based.variants.base.utils.constants import ENABLE_LOOP_DETECTION, \
    ENABLE_XOR_DETECTION
from powl.discovery.partial_order_based.variants.base.utils.mapping import apply_node_mapping_on_single_graph, \
    find_self_loops
from powl.discovery.partial_order_based.variants.base.utils.node_grouping import NodeGrouping
from powl.discovery.partial_order_based.variants.base.utils.xor_detection import XORMiner


def apply_mining_algorithm_recursively(node):
    if isinstance(node, ActivityInstance):
        if ENABLE_LOOP_DETECTION:
            return ActivityInstance(node.label, 1)
        else:
            return node
    elif isinstance(node, Skip):
        node_element = apply_mining_algorithm_recursively(node.element)
        return Skip.create(node_element)
    elif isinstance(node, Graph):
        return _mine([node])
    elif isinstance(node, LOOP):
        body = apply_mining_algorithm_recursively(node.body)
        redo = apply_mining_algorithm_recursively(node.redo)
        return LOOP(body, redo)
    elif isinstance(node, XOR):
        new_children = {apply_mining_algorithm_recursively(child) for child in node.children}
        return XOR(frozenset(new_children))
    else:
        raise TypeError('Unsupported node type')


def _mine(orders):

    if len(orders) < 1:
        raise ValueError("Input list of partial orders is empty!")

    all_activity_labels = set()

    for graph in orders:
        for node in graph.nodes:
            all_activity_labels.update(get_leaves(node))

    if ENABLE_LOOP_DETECTION:
        if len(all_activity_labels) == 1:
            activity_label = all_activity_labels.pop()
            activity = ActivityInstance(activity_label, 1)
            if any(len(order.nodes) > 1 for order in orders):
                return SelfLoop.create(activity)
            else:
                return activity

    if ENABLE_XOR_DETECTION:
        xor_clusters = XORMiner.find_disjoint_activities(orders, all_activity_labels)

        label_mapping = {}
        if xor_clusters is not None:

            for cluster in xor_clusters:

                sub_models = []
                for group in cluster:
                    projected_log = XORMiner.project_partial_orders_on_groups(orders, list(group))
                    sub_models.append(_mine(projected_log))
                model = XOR(children=frozenset(sub_models))
                for group in cluster:
                    for activity_label in group:
                        if activity_label in label_mapping.keys():
                            raise ValueError("Duplicate activity label")
                        label_mapping[activity_label] = model

            orders = XORMiner.apply_mapping(orders, label_mapping)


    mapping_skips, new_nodes_counter = NodeGrouping.find_groups(orders)
    if ENABLE_LOOP_DETECTION:
        node_mapping = find_self_loops(mapping_skips, new_nodes_counter)
    else:
        node_mapping = mapping_skips
    orders = [apply_node_mapping_on_single_graph(g, node_mapping) for g in orders]

    if len(orders) == 1:
        order = orders[0]
    else:
        order = combine_orders(orders)


    if len(order.nodes) == 0:
        return ActivityInstance(None, 1)

    if len(order.nodes) == 1:
        return list(order.nodes)[0]

    return order


def apply(partial_orders):
    order = _mine(partial_orders)
    powl = generate_powl(order)
    return powl


from enum import Enum, auto
from collections import Counter
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL


class FilteringType(Enum):
    DYNAMIC = auto()
    DECREASING_FACTOR = auto()
    DFG_FREQUENCY = auto()


DEFAULT_FILTERING_TYPE = FilteringType.DECREASING_FACTOR
FILTERING_THRESHOLD = "weight_factor_filtering_threshold"
FILTERING_TYPE = "filtering_type"


def filter_most_frequent_variants(log):
    to_remove_freq = min([freq for var, freq in log.items()])
    new_log = Counter()
    for var, freq in log.items():
        if freq == to_remove_freq:
            continue
        new_log[var] = freq

    return IMDataStructureUVCL(new_log)


def filter_most_frequent_variants_with_decreasing_factor(log, decreasing_factor):
    sorted_variants = sorted(log, key=log.get, reverse=True)
    new_log = Counter()

    already_added_sum = 0
    prev_var_count = -1

    for variant in sorted_variants:
        frequency = log[variant]
        if already_added_sum == 0 or frequency > decreasing_factor * prev_var_count:
            new_log[variant] = frequency
            already_added_sum = already_added_sum + frequency
            prev_var_count = frequency
        else:
            break

    return IMDataStructureUVCL(new_log)



from abc import ABC
from typing import Optional, Any, Dict, Generic

from pm4py.algo.discovery.inductive.cuts.xor import ExclusiveChoiceCut, ExclusiveChoiceCutUVCL, T
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from powl.objects.obj import OperatorPOWL
from pm4py.objects.process_tree.obj import Operator


class POWLExclusiveChoiceCut(ExclusiveChoiceCut, ABC, Generic[T]):

    @classmethod
    def operator(cls, parameters: Optional[Dict[str, Any]] = None) -> OperatorPOWL:
        return OperatorPOWL(Operator.XOR, [])


class POWLExclusiveChoiceCutUVCL(ExclusiveChoiceCutUVCL, POWLExclusiveChoiceCut[IMDataStructureUVCL], ABC):
    pass



import warnings
from bisect import bisect_left
from collections import defaultdict
from typing import List, Any, Set
import pandas as pd

from powl.discovery.partial_order_based.utils.constants import VARIANT_FREQUENCY_KEY
from powl.discovery.partial_order_based.utils.simplified_objects import ActivityInstance, Graph
from powl.general_utils.time_utils import should_parse_column_as_date


def generate_interval_df_fifo(
        df: pd.DataFrame,
        case_id_col: str, activity_col: str, ordering_col: str, lifecycle_col: str or None,
        start_transitions: Set[str], complete_transitions: Set[str]
) -> pd.DataFrame:

    cols_to_keep = [case_id_col, activity_col, ordering_col]
    if lifecycle_col:
        cols_to_keep.append(lifecycle_col)
    df_filtered = df[cols_to_keep].copy()

    if lifecycle_col:
        df_filtered = df_filtered.sort_values([case_id_col, ordering_col])

        activity_col_idx = df_filtered.columns.get_loc(activity_col)
        timestamp_col_idx = df_filtered.columns.get_loc(ordering_col)
        lifecycle_col_idx = df_filtered.columns.get_loc(lifecycle_col)

        all_intervals = []
        grouped_by_case = df_filtered.groupby(case_id_col, sort=False)

        for case_id, case_df in grouped_by_case:
            pending_starts = defaultdict(list)
            activity_instance_counter = defaultdict(int)

            for event_tuple in case_df.itertuples(index=False, name=None):
                activity_name = event_tuple[activity_col_idx]
                timestamp = event_tuple[timestamp_col_idx]
                lifecycle = event_tuple[lifecycle_col_idx]

                if lifecycle in start_transitions:
                    pending_starts[activity_name].append(timestamp)

                elif lifecycle in complete_transitions:
                    if pending_starts[activity_name]:
                        start_timestamp = pending_starts[activity_name].pop(0)
                    else:
                        start_timestamp = timestamp
                    activity_instance_counter[activity_name] += 1
                    all_intervals.append({
                        case_id_col: case_id,
                        'activity': ActivityInstance(activity_name, activity_instance_counter[activity_name]),
                        'start_timestamp': start_timestamp,
                        'end_timestamp': timestamp
                    })

        interval_df = pd.DataFrame(all_intervals) if all_intervals else pd.DataFrame()

    else:
        # No lifecycle: each event is an atomic interval
        df_filtered = df_filtered.sort_values([case_id_col, ordering_col])
        group_cols = [case_id_col, activity_col]
        df_filtered['instance_id'] = df_filtered.groupby(group_cols).cumcount()

        df_filtered['activity'] = [ActivityInstance(act, inst + 1) for act, inst in
                                   zip(df_filtered[activity_col], df_filtered['instance_id'])]
        df_filtered['start_timestamp'] = df_filtered[ordering_col]
        df_filtered['end_timestamp'] = df_filtered[ordering_col]

        interval_df = df_filtered[[case_id_col, 'activity', 'start_timestamp', 'end_timestamp']]

    if interval_df.empty:
        return pd.DataFrame()

    # Sort for consistent output and to aid the next processing step
    interval_df = interval_df.sort_values([case_id_col, 'start_timestamp', 'end_timestamp']).reset_index(drop=True)
    interval_df['event_instance_id'] = interval_df.index
    print(f"Successfully created {len(interval_df)} activity intervals using FIFO logic.")
    return interval_df



def apply(
    df: pd.DataFrame,
    case_id_col: str,
    activity_col: str,
    ordering_col: str,
    lifecycle_col: str or None,
    start_transitions: Set[str],
    complete_transitions: Set[str],
) -> List[Any]:

    if not isinstance(df, pd.DataFrame):
        raise TypeError("Input 'log' must be a Pandas DataFrame.")

    for col in [activity_col, ordering_col, case_id_col]:
        if col not in df.columns:
            raise ValueError(f"Column '{col}' not found in the event table!")

    if lifecycle_col:
        if lifecycle_col in df.columns:
            complete_log = df[df['lifecycle:transition'].isin(complete_transitions)]
            if len(complete_log) == 0:
                lifecycle_col = None
                warnings.warn(message=f'The event log does not contain any completion events! All events will be considered during discovery!', category=UserWarning)
        else:
            lifecycle_col = None
            warnings.warn(f'The event log does not contain any attribute with the lifecycle key "{lifecycle_col}"! All events will be considered during discovery!', category=UserWarning)

    if should_parse_column_as_date(df, ordering_col):
        df[ordering_col] = pd.to_datetime(df[ordering_col])

    interval_df = generate_interval_df_fifo(
        df, case_id_col, activity_col, ordering_col, lifecycle_col,
        start_transitions, complete_transitions
    )

    if interval_df.empty:
        raise Exception("Interval DataFrame is empty, no variants to generate.")

    variants_key_to_frequency = defaultdict(int)
    interval_df = interval_df.sort_values([case_id_col, 'start_timestamp', 'end_timestamp']).reset_index(drop=True)
    grouped_intervals = interval_df.groupby(case_id_col, sort=False)

    for case_id, trace_df in grouped_intervals:
        trace_activities_multiset = frozenset(trace_df['activity'].tolist())

        trace_events = list(
            trace_df[['activity', 'start_timestamp', 'end_timestamp']].itertuples(index=False, name=None))
        events_sorted_by_end = sorted(trace_events, key=lambda x: x[2])
        end_timestamps = [event[2] for event in events_sorted_by_end]

        edges = []
        for act_j, start_j, _ in trace_events:
            pos = bisect_left(end_timestamps, start_j)
            for i in range(pos):
                act_i = events_sorted_by_end[i][0]
                if act_i != act_j:
                    edges.append((act_i, act_j))

        trace_edges = frozenset(edges)
        variant_key = (trace_activities_multiset, trace_edges)
        variants_key_to_frequency[variant_key] += 1

    print(f"Found {len(variants_key_to_frequency)} unique variants.")

    output_list = [
        Graph(variant_key[0], variant_key[1], {VARIANT_FREQUENCY_KEY: freq})
        for variant_key, freq in variants_key_to_frequency.items()
    ]

    output_list.sort(
        key=lambda x: x.additional_information[VARIANT_FREQUENCY_KEY],
        reverse=True
    )

    return output_list




from typing import List, Optional, Dict, Any, Tuple, Type

from powl.discovery.total_order_based.inductive.cuts.concurrency import POWLConcurrencyCutUVCL
from powl.discovery.total_order_based.inductive.cuts.factory import S, T, CutFactory
from powl.discovery.total_order_based.inductive.cuts.loop import POWLLoopCutUVCL
from powl.discovery.total_order_based.inductive.cuts.sequence import POWLStrictSequenceCutUVCL
from powl.discovery.total_order_based.inductive.cuts.xor import POWLExclusiveChoiceCutUVCL
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from powl.discovery.total_order_based.inductive.variants.brute_force.bf_partial_order_cut import BruteForcePartialOrderCutUVCL
from powl.objects.obj import POWL


class CutFactoryPOWLBruteForce(CutFactory):

    @classmethod
    def get_cuts(cls, obj: T, parameters: Optional[Dict[str, Any]] = None) -> List[Type[S]]:
        if type(obj) is IMDataStructureUVCL:
            return [POWLExclusiveChoiceCutUVCL, POWLStrictSequenceCutUVCL, POWLConcurrencyCutUVCL, POWLLoopCutUVCL,
                    BruteForcePartialOrderCutUVCL]
        return list()

    @classmethod
    def find_cut(cls, obj: IMDataStructureUVCL, parameters: Optional[Dict[str, Any]] = None) -> Optional[
            Tuple[POWL, List[T]]]:
        for c in CutFactoryPOWLBruteForce.get_cuts(obj):
            r = c.apply(obj, parameters)
            if r is not None:
                return r
        return None



import pandas as pd


def should_parse_column_as_date(df, column):
    """Check the first non-empty value in a column to decide datetime parsing."""
    try:
        first_val = df[column].dropna().astype(str).iloc[0]
        pd.to_datetime(first_val, errors='raise')
        return True
    except Exception:
        return False


from powl.discovery.total_order_based.inductive.base_case.abc import BaseCase
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from typing import Optional, Dict, Any

from powl.objects.obj import Transition


class SingleActivityBaseCaseUVCL(BaseCase[IMDataStructureUVCL]):
    @classmethod
    def holds(cls, obj=IMDataStructureUVCL, parameters: Optional[Dict[str, Any]] = None) -> bool:
        if len(obj.data_structure.keys()) != 1:
            return False
        if len(list(obj.data_structure.keys())[0]) > 1:
            return False
        return True

    @classmethod
    def leaf(cls, obj=IMDataStructureUVCL, parameters: Optional[Dict[str, Any]] = None) -> Transition:
        for t in obj.data_structure:
            return Transition(label=t[0])




import copy
import importlib.resources
import os

from powl.objects.obj import POWL
from enum import Enum
import tempfile
import graphviz
from graphviz import Digraph
from pm4py.util import constants
from pm4py.objects.bpmn.obj import BPMN
from pm4py.objects.bpmn.util.sorting import get_sorted_nodes_edges
from powl.conversion.variants.to_petri_net import apply as powl_to_pn
from pm4py.objects.bpmn.util import reduction


class Parameters(Enum):
    FORMAT = "format"
    RANKDIR = "rankdir"
    BGCOLOR = "bgcolor"


FREQUENCY_TAG_IMAGES = True
SPLIT_EXCLUSIVE_GATEWAYS = True
min_width = "1.5"  # Set the minimum width in inches
min_height = "0.5"
node_color = "#f2f2f2"
silent_node_color = "black"


class SplitExclusiveGateway(BPMN.ExclusiveGateway):
    pass


class JoinExclusiveGateway(BPMN.ExclusiveGateway):
    pass


def can_connect_without_xor_violation(input_nodes, output_nodes):
    for in_node in input_nodes:
        for out_node in output_nodes:
            if isinstance(in_node, BPMN.ExclusiveGateway) and len(output_nodes) > 1:
                return False
            if isinstance(out_node, BPMN.ExclusiveGateway) and len(input_nodes) > 1:
                return False

    return True


def remove_all_occurrences_from_list(item, items):
    return [other for other in items if other is not item]


def simplify_and_gateways(nodes, edges):
    for n in list(nodes):
        if isinstance(n, BPMN.ParallelGateway):
            input_nodes = [edge[0] for edge in edges if edge[1] == n]
            output_nodes = [edge[1] for edge in edges if edge[0] == n]
            if can_connect_without_xor_violation(input_nodes, output_nodes):
                nodes = remove_all_occurrences_from_list(n, nodes)
                edges = [edge for edge in edges if n not in edge]

                for in_node in input_nodes:
                    for out_node in output_nodes:
                        edges.append((in_node, out_node))
                return simplify_and_gateways(nodes, edges)

    return nodes, edges


def add_node(n, viz):
    n_id = str(id(n))
    if isinstance(n, FrequencyTask):
        font_size = '18'
        if FREQUENCY_TAG_IMAGES:
            if n.skippable:
                if n.selfloop:
                    with importlib.resources.path("powl.visualization.powl.variants.icons",
                                                  "skip-loop-tag.svg") as gimg:
                        image = str(gimg)
                        viz.node(n_id, label='\n' + n.activity, imagepos='tr',
                                 image=image, style='filled', fillcolor=node_color,
                                 shape='box', width=min_width, fontsize=font_size)
                else:
                    with importlib.resources.path("powl.visualization.powl.variants.icons",
                                                  "skip-tag.svg") as gimg:
                        skip_image = str(gimg)
                        viz.node(n_id, label="\n" + n.activity, imagepos='tr',
                                 image=skip_image, style='filled', fillcolor=node_color,
                                 shape='box', width=min_width, fontsize=font_size)
            else:
                if n.selfloop:
                    with importlib.resources.path("powl.visualization.powl.variants.icons",
                                                  "loop-tag.svg") as gimg:
                        loop_image = str(gimg)
                        viz.node(n_id, label="\n" + n.activity, imagepos='tr',
                                 image=loop_image, style='filled', fillcolor=node_color,
                                 shape='box', width=min_width, fontsize=font_size)
                else:
                    viz.node(n_id, label=n.activity, style='filled', fillcolor=node_color,
                             shape='box', width=min_width, fontsize=font_size)
        else:
            if n.skippable:
                viz.node(n_id, shape="box", label=n.activity, fontsize=font_size,
                         style="dashed", width=min_width)
            else:
                viz.node(n_id, shape="box", label=n.activity, fontsize=font_size,
                         width=min_width)
            if n.selfloop:
                viz.edge(n_id, n_id)

    elif isinstance(n, BPMN.StartEvent):
        with importlib.resources.path("powl.visualization.powl.variants.icons", "play.svg") as gimg:
            start_image = str(gimg)
            viz.node(n_id, image=start_image, label="", shape="none", width='0.35',
                     height='0.35', fixedsize="true", style='filled', fillcolor=node_color)
    elif isinstance(n, BPMN.EndEvent):
        with importlib.resources.path("powl.visualization.powl.variants.icons", "end.svg") as gimg:
            end_image = str(gimg)
            viz.node(n_id, image=end_image, label="", shape="none", width='0.35',
                     height='0.35', fixedsize="true", style='filled', fillcolor=node_color)
    elif isinstance(n, BPMN.ParallelGateway):
        viz.node(n_id, label="", shape="square", style="filled", fillcolor=silent_node_color, width='0.3',
                 height='0.3')
    elif isinstance(n, SplitExclusiveGateway):
        with importlib.resources.path("powl.visualization.powl.variants.icons", "gate.svg") as gimg:
            xor_image = str(gimg)
            viz.node(n_id, label="", shape="diamond", style="filled", fillcolor="lightgreen",
                     width='0.4', height='0.4', fixedsize="true", image=xor_image)
    elif isinstance(n, JoinExclusiveGateway):
        with importlib.resources.path("powl.visualization.powl.variants.icons", "gate.svg") as gimg:
            xor_image = str(gimg)
            viz.node(n_id, label="", shape="diamond", style="filled", fillcolor="orange",
                     width='0.4', height='0.4', fixedsize="true", image=xor_image)
    elif isinstance(n, BPMN.ExclusiveGateway):
        with importlib.resources.path("powl.visualization.powl.variants.icons", "gate.svg") as gimg:
            xor_image = str(gimg)
        viz.node(n_id, label="", shape="diamond", style="filled", fillcolor=node_color,
                 width='0.4', height='0.4', fixedsize="true", image=xor_image)
    else:
        raise Exception("Unexpected instance of class " + str(type(n)) + "!")


def split_xor_gateways(nodes, edges):
    for node in list(nodes):
        if isinstance(node, BPMN.ExclusiveGateway):
            preset = [e[0] for e in edges if e[1] is node]
            postset = [e[1] for e in edges if e[0] is node]

            if len(preset) == 1 and len(postset) == 1:
                edges = remove_all_occurrences_from_list((preset[0], node), edges)
                edges = remove_all_occurrences_from_list((node, postset[0]), edges)
                edges.append((preset[0], postset[0]))

            elif len(preset) == 1 and len(postset) > 1:
                node.__class__ = SplitExclusiveGateway

            elif len(preset) > 1 and len(postset) == 1:
                node.__class__ = JoinExclusiveGateway

            elif len(preset) > 1 and len(postset) > 1:
                join_node = copy.deepcopy(node)
                join_node.__class__ = JoinExclusiveGateway
                nodes.append(join_node)

                split_node = copy.deepcopy(node)
                split_node.__class__ = SplitExclusiveGateway
                nodes.append(split_node)
                nodes = remove_all_occurrences_from_list(node, nodes)

                for idx, edge in enumerate(edges):
                    if edge[1] is node:
                        edges[idx] = (edge[0], join_node)
                    if edge[0] is node:
                        edges[idx] = (split_node, edge[1])

                edges.append((join_node, split_node))
    return nodes, edges


def apply(powl: POWL) -> graphviz.Digraph:
    pn_2, init_2, final_2 = powl_to_pn(powl, parameters={"flatten_frequency_tags":False})
    bpmn_graph = to_bpmn(pn_2, init_2, final_2)

    nodes, edges = get_sorted_nodes_edges(bpmn_graph)

    nodes, edges = simplify_and_gateways(nodes, edges)

    rankdir = "LR"
    bgcolor = constants.DEFAULT_BGCOLOR

    filename = tempfile.NamedTemporaryFile(suffix='.gv')
    viz = Digraph("", filename=filename.name, engine='dot', graph_attr={'bgcolor': bgcolor})
    viz.graph_attr['rankdir'] = rankdir

    gateway_edges = {}

    for e in edges:
        if isinstance(e[0], BPMN.ParallelGateway) or isinstance(e[0], BPMN.ParallelGateway):
            if e[0] not in gateway_edges:
                gateway_edges[e[0]] = {'in': [], 'out': []}
            gateway_edges[e[0]]['out'].append(e[1])
            continue
        if isinstance(e[1], BPMN.ParallelGateway) or isinstance(e[1], BPMN.ParallelGateway):
            if e[1] not in gateway_edges:
                gateway_edges[e[1]] = {'in': [], 'out': []}
            gateway_edges[e[1]]['in'].append(e[0])
            continue

    if SPLIT_EXCLUSIVE_GATEWAYS:
        nodes, edges = split_xor_gateways(nodes, edges)

    for node in nodes:
        add_node(node, viz)
    add_concurrent_subgraphs(viz, find_concurrent_groups(nodes, edges))

    for e in edges:
        n_id_1 = str(id(e[0]))
        n_id_2 = str(id(e[1]))

        viz.edge(n_id_1, n_id_2)

    viz.attr(overlap='false')

    viz.format = "svg"

    return viz


class FrequencyTask(BPMN.Task):
    def __init__(self, name, properties, id="", in_arcs=None, out_arcs=None, process=None):
        super().__init__(id=id, name=name, in_arcs=in_arcs, out_arcs=out_arcs, process=process)
        self.activity = properties["activity"]
        self.skippable = properties["skippable"]
        self.selfloop = properties["selfloop"]
        script_dir = os.path.dirname(os.path.realpath(__file__))
        self.image = os.path.join(script_dir, "xor2.png")


def to_bpmn(net, im, fm):
    """
    Converts an accepting Petri net into a BPMN diagram

    Parameters
    --------------
    net
        Petri net
    im
        initial marking
    fm
        final marking

    Returns
    --------------
    bpmn_graph
        BPMN diagram
    """

    bpmn_graph = BPMN()

    entering_dictio = {}
    exiting_dictio = {}

    for place in net.places:
        node = BPMN.ExclusiveGateway()
        bpmn_graph.add_node(node)
        entering_dictio[place] = node
        exiting_dictio[place] = node

    for trans in net.transitions:
        if trans.label is None:
            if len(trans.in_arcs) > 1:
                node = BPMN.ParallelGateway(gateway_direction=BPMN.Gateway.Direction.CONVERGING)
            elif len(trans.out_arcs) > 1:
                node = BPMN.ParallelGateway(gateway_direction=BPMN.Gateway.Direction.DIVERGING)
            else:
                node = BPMN.ExclusiveGateway(gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED)
            bpmn_graph.add_node(node)
            entering_dictio[trans] = node
            exiting_dictio[trans] = node
        else:
            if len(trans.in_arcs) > 1:
                entering_node = BPMN.ParallelGateway(gateway_direction=BPMN.Gateway.Direction.CONVERGING)
            else:
                entering_node = BPMN.ExclusiveGateway(gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED)

            if len(trans.out_arcs) > 1:
                exiting_node = BPMN.ParallelGateway(gateway_direction=BPMN.Gateway.Direction.DIVERGING)
            else:
                exiting_node = BPMN.ExclusiveGateway(gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED)

            task = FrequencyTask(name=trans.label, properties=trans.properties)
            bpmn_graph.add_node(task)

            bpmn_graph.add_flow(BPMN.Flow(entering_node, task))
            bpmn_graph.add_flow(BPMN.Flow(task, exiting_node))

            entering_dictio[trans] = entering_node
            exiting_dictio[trans] = exiting_node

    for arc in net.arcs:
        bpmn_graph.add_flow(BPMN.Flow(exiting_dictio[arc.source], entering_dictio[arc.target]))

    start_node = BPMN.StartEvent(name="start", isInterrupting=True)
    end_node = BPMN.NormalEndEvent(name="end")
    bpmn_graph.add_node(start_node)
    bpmn_graph.add_node(end_node)
    for place in im:
        bpmn_graph.add_flow(BPMN.Flow(start_node, entering_dictio[place]))
    for place in fm:
        bpmn_graph.add_flow(BPMN.Flow(exiting_dictio[place], end_node))

    bpmn_graph = reduction.apply(bpmn_graph)

    for node in bpmn_graph.get_nodes():
        node.set_process(bpmn_graph.get_process_id())

    for edge in bpmn_graph.get_flows():
        edge.set_process(bpmn_graph.get_process_id())

    return bpmn_graph


def find_concurrent_groups(nodes, edges):
    predecessors = {n: [] for n in nodes}
    successors = {n: [] for n in nodes}
    for src, dst in edges:
        successors[src].append(dst)
        predecessors[dst].append(src)

    groups = {}
    for node, preds in predecessors.items():
        if len(preds) == 1:
            pred_id = str(id(preds[0]))
            if pred_id not in groups:
                groups[pred_id] = []
            groups[pred_id].append(node)

    return [group for group in groups.values() if len(group) > 1]


def add_concurrent_subgraphs(graph, concurrent_elements):
    for group in concurrent_elements:
        with graph.subgraph() as s:
            s.attr(rank='same')
            for n in group:
                s.node(str(id(n)))



from collections import Counter
from multiprocessing import Pool, Manager
from typing import Optional, Tuple, List, Any, Dict

from pm4py.algo.discovery.inductive.fall_through.activity_concurrent import ActivityConcurrentUVCL
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from powl.objects.obj import StrictPartialOrder


class POWLActivityConcurrentUVCL(ActivityConcurrentUVCL):

    @classmethod
    def apply(cls, obj: IMDataStructureUVCL, pool: Pool = None, manager: Manager = None,
              parameters: Optional[Dict[str, Any]] = None) -> Optional[
        Tuple[StrictPartialOrder, List[IMDataStructureUVCL]]]:
        candidate = cls._get_candidate(obj, pool, manager, parameters)
        if candidate is None:
            return None
        log = obj.data_structure
        l_a = Counter()
        l_other = Counter()
        for t in log:
            l_a.update({tuple(filter(lambda e: e == candidate, t)): log[t]})
            l_other.update({tuple(filter(lambda e: e != candidate, t)): log[t]})
        return StrictPartialOrder([]), [IMDataStructureUVCL(l_a), IMDataStructureUVCL(l_other)]



from abc import ABC
from typing import Generic, Optional, Dict, Any

from powl.discovery.total_order_based.inductive.base_case.abc import BaseCase, T
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from powl.objects.obj import SilentTransition


class EmptyLogBaseCase(BaseCase[T], ABC, Generic[T]):

    @classmethod
    def leaf(cls, obj=T, parameters: Optional[Dict[str, Any]] = None) -> SilentTransition:
        return SilentTransition()


class EmptyLogBaseCaseUVCL(EmptyLogBaseCase[IMDataStructureUVCL]):

    @classmethod
    def holds(cls, obj=IMDataStructureUVCL, parameters: Optional[Dict[str, Any]] = None) -> bool:
        return len(obj.data_structure) == 0




from typing import TypeVar, Optional, Dict, Any, Type, List as TList

from powl.discovery.total_order_based.inductive.base_case.abc import BaseCase
from powl.discovery.total_order_based.inductive.base_case.empty_log import EmptyLogBaseCaseUVCL
from powl.discovery.total_order_based.inductive.base_case.single_activity import SingleActivityBaseCaseUVCL
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructure, IMDataStructureUVCL

from powl.objects.obj import POWL

T = TypeVar('T', bound=IMDataStructure)
S = TypeVar('S', bound=BaseCase)


class BaseCaseFactory:

    @classmethod
    def get_base_cases(cls, obj: T, parameters: Optional[Dict[str, Any]] = None) -> TList[Type[S]]:
        if type(obj) is IMDataStructureUVCL:
            return [EmptyLogBaseCaseUVCL, SingleActivityBaseCaseUVCL]
        return []

    @classmethod
    def apply_base_cases(cls, obj: T, parameters: Optional[Dict[str, Any]] = None) -> Optional[POWL]:
        for b in BaseCaseFactory.get_base_cases(obj):
            r = b.apply(obj, parameters)
            if r is not None:
                return r
        return None



from abc import ABC, abstractmethod
from typing import Union, TypeVar, Generic, Optional, Dict, Any

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructure
from powl.objects.obj import POWL

T = TypeVar('T', bound=Union[IMDataStructure])


class BaseCase(ABC, Generic[T]):

    @classmethod
    def apply(cls, obj=T, parameters: Optional[Dict[str, Any]] = None) -> Optional[POWL]:
        return cls.leaf(obj, parameters) if cls.holds(obj, parameters) else None

    @classmethod
    @abstractmethod
    def holds(cls, obj=T, parameters: Optional[Dict[str, Any]] = None) -> bool:
        pass

    @classmethod
    @abstractmethod
    def leaf(cls, obj=T, parameters: Optional[Dict[str, Any]] = None) -> POWL:
        pass



from powl.main import import_event_log, discover, view, view_net, save_visualization, save_visualization_net, convert_to_petri_net, convert_to_bpmn, discover_from_partially_ordered_log, import_ocel, discover_petri_net_from_ocel

__name__ = 'powl'
__version__ = '2.0.2'



from typing import Set

from powl.objects.obj import SilentTransition, Transition, OperatorPOWL, StrictPartialOrder, Operator

ENABLE_DUPLICATION = True

class XOR:
    def __init__(self, children: frozenset):
        if not isinstance(children, frozenset):
            raise TypeError("Children must be provided as a frozenset.")
        if len(children) < 1:
            raise ValueError("XOR must have at least one child.")
        self.children = children

    def __repr__(self):
        return f"XOR({', '.join(repr(child) for child in sorted(self.children))})"

    def __eq__(self, other):
        if isinstance(other, XOR):
            return self.children == other.children
        return False

    def __hash__(self):
        return hash(('XOR', self.children))

    def __lt__(self, other):
        if isinstance(other, XOR):
            return sorted(self.children) < sorted(other.children)
        elif isinstance(other, LOOP) or isinstance(other, Graph): # ActivityInstance < XOR < Loop < Graph
            return True
        elif isinstance(other, ActivityInstance):
            return False
        else:
            return NotImplemented

class Skip(XOR):
    _allow_init = False

    def __init__(self, element):
        if not Skip._allow_init:
            raise RuntimeError("You must use create() to create this object!")
        super().__init__(frozenset([element, ActivityInstance(None, 1)]))
        self.element = element

    @classmethod
    def create(cls, element):
        if isinstance(element, Skip) or isinstance(element, SkipSelfLoop):
            instance = element
        elif isinstance(element, SelfLoop):
            return SkipSelfLoop.create(element.element)
        else:
            cls._allow_init = True
            instance = cls(element)
            cls._allow_init = False

        return instance

    @property
    def allow_init(self):
        return self._allow_init


class LOOP:
    def __init__(self, body, redo):
        self.body = body
        self.redo = redo

    def __repr__(self):
        return f"LOOP(body={repr(self.body)}, redo={repr(self.redo)})"

    def __eq__(self, other):
        if isinstance(other, LOOP):
            return self.body == other.body and self.redo == other.redo
        return False

    def __hash__(self):
        return hash(('LOOP', self.body, self.redo))

    def __lt__(self, other):
        if isinstance(other, LOOP):
            return (self.body, self.redo) < (other.body, other.redo)
        elif isinstance(other, Graph): # ActivityInstance < XOR < Loop < Graph
            return True
        elif isinstance(other, XOR) or isinstance(other, ActivityInstance):
            return False
        else:
            return NotImplemented


class SelfLoop(LOOP):
    _allow_init = False

    def __init__(self, element):
        if not SelfLoop._allow_init:
            raise RuntimeError("You must use create() to create this object!")
        super().__init__(element, ActivityInstance(None, 1))
        self.element = element


    @classmethod
    def create(cls, element):
        if isinstance(element, SelfLoop):
            return element
        elif isinstance(element, Skip) or isinstance(element, SkipSelfLoop):
            return SkipSelfLoop.create(element.element)
        else:
            cls._allow_init = True
            instance = cls(element)
            cls._allow_init = False
            return instance


class SkipSelfLoop(LOOP):
    def __init__(self, element):
        if isinstance(element, SkipSelfLoop) or isinstance(element, SelfLoop) or isinstance(element, Skip):
            element = element.element
        super().__init__(ActivityInstance(None, 1), element)
        self.element = element

    @classmethod
    def create(cls, element):
        return cls(element)


class ActivityInstance:
    def __init__(self, label: str|None, number: int):
        if not ENABLE_DUPLICATION:
            number = 1
        if number < 1:
            raise ValueError("Activity number must be at least 1.")
        self.label = label
        self.number = number

    def __repr__(self):
        if self.number == 1:
            return f"{self.label}"
        return f"({self.label}, {self.number})"

    def __eq__(self, other):
        if isinstance(other, ActivityInstance):
            return self.label == other.label and self.number == other.number
        return False

    def __hash__(self):
        return hash((self.label, self.number))

    def __lt__(self, other):
        if isinstance(other, ActivityInstance):
            if self.label and not other.label:
                return False
            elif not self.label and other.label:
                return True
            return (self.label, self.number) < (other.label, other.number)
        # ActivityInstance < XOR < Loop < Graph
        elif isinstance(other, Graph) or isinstance(other, XOR) or isinstance(other, LOOP):
            return True
        else:
            return NotImplemented


class Graph:
    def __init__(self, nodes: frozenset, edges: frozenset, additional_information=None):
        if additional_information is None:
            additional_information = {}
        if not isinstance(nodes, frozenset):
            raise TypeError("Nodes must be a frozenset.")
        if not isinstance(edges, frozenset):
            raise TypeError("Edges must be a frozenset.")
        for edge in edges:
            if not (isinstance(edge, tuple) and len(edge) == 2):
                raise ValueError(f"Each edge must be a (source, target) tuple, found: {edge}")
            if edge[0] not in nodes or edge[1] not in nodes:
                raise ValueError(f"Edge {edge} refers to nodes not in the node set.")

        self.nodes = nodes
        self.edges = edges
        self.additional_information = additional_information if additional_information else {}

    def __repr__(self):
        nodes_repr = ', '.join(sorted(map(repr, self.nodes)))
        edges_repr = ', '.join(f"{repr(src)}->{repr(tgt)}" for src, tgt in sorted(self.edges))
        return f"Graph(Nodes: {{{nodes_repr}}}, Edges: {{{edges_repr}}}, {self.additional_information})"

    def __eq__(self, other):
        if isinstance(other, Graph):
            return self.nodes == other.nodes and self.edges == other.edges
        return False

    def __hash__(self):
        return hash(('Graph', self.nodes, self.edges))

    def __lt__(self, other):
        if isinstance(other, Graph):
            return (sorted(self.nodes), sorted(self.edges)) < (sorted(other.nodes), sorted(other.edges))
        # ActivityInstance < XOR < Loop < Graph
        elif isinstance(other, XOR) or isinstance(other, ActivityInstance) or isinstance(other, LOOP):
            return False
        else:
            return NotImplemented


def get_leaves(node) -> Set[str]:
    if isinstance(node, ActivityInstance):
        return {node.label} if node.label else set()
    if isinstance(node, Graph):
        children = node.nodes
    elif isinstance(node, XOR):
        children = node.children
    elif isinstance(node, LOOP):
        children = [node.body, node.redo]
    else:
        raise TypeError
    res = set()
    for child in children:
        res.update(get_leaves(child))
    return res


def _simplified_model_to_powl(model, add_instance_number = False):
    if isinstance(model, ActivityInstance):
        if not model.label:
            return SilentTransition()
        if add_instance_number:
            label = f"({model.label}, {model.number})"
        else:
            label = model.label
        return Transition(label=label)
    elif isinstance(model, XOR):
        return OperatorPOWL(operator=Operator.XOR, children=[_simplified_model_to_powl(child) for child in model.children])
    elif isinstance(model, LOOP):
        return OperatorPOWL(operator=Operator.LOOP, children=[_simplified_model_to_powl(model.body), _simplified_model_to_powl(model.redo)])
    elif not isinstance(model, Graph):
        raise NotImplementedError


    po = StrictPartialOrder([])
    submodels = model.nodes
    edges = model.edges

    powl_map = {}
    for submodel in submodels:
        powl_child = _simplified_model_to_powl(submodel)
        powl_map[submodel] = powl_child
        po.order.add_node(powl_child)

    for m1, m2 in edges:
        po.order.add_edge(powl_map[m1], powl_map[m2])

    len_all = len(po.order.nodes)

    start_len = len(po.order.get_start_nodes())
    if start_len > 1 and start_len != len_all:
        start = SilentTransition()
        po.order.add_node(start)
        for node in set(po.order.nodes) - {start}:
            po.order.add_edge(start, node)

    end_len = len(po.order.get_end_nodes())
    if end_len > 1 and end_len != len_all:
        end = SilentTransition()
        po.order.add_node(end)
        for node in set(po.order.nodes) - {end}:
            po.order.add_edge(node, end)

    if not po.order.is_irreflexive():
        raise ValueError('Not irreflexive!')

    if not po.order.is_transitive():
        raise ValueError('Not transitive!')

    return po


def generate_powl(model, add_instance_number = False):
    powl = _simplified_model_to_powl(model, add_instance_number = add_instance_number)
    return powl.simplify()


import importlib

from pm4py.util import exec_utils
from enum import Enum
import tempfile
from graphviz import Digraph
from typing import Optional, Dict, Any
from pm4py.objects.bpmn.obj import BPMN
from pm4py.util import constants
import graphviz
from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave


SPLIT_LABELS = False


class Parameters(Enum):
    FORMAT = "format"
    RANKDIR = "rankdir"
    FONT_SIZE = "font_size"
    BGCOLOR = "bgcolor"


def get_label(name):
    parts = name.split(" ")
    if SPLIT_LABELS and len(parts) >= 3:
        return " ".join(parts[:2]) + "\n" + " ".join(parts[2:])
    else:
        return name


def apply(bpmn_graph: BPMN, parameters: Optional[Dict[Any, Any]] = None) -> graphviz.Digraph:
    """
    Visualize a BPMN graph

    Parameters
    -------------
    bpmn_graph
        BPMN graph
    parameters
        Parameters of the visualization, including:
         - Parameters.FORMAT: the format of the visualization
         - Parameters.RANKDIR: the direction of the representation (default: LR)

    Returns
    ------------
    gviz
        Graphviz representation
    """
    if parameters is None:
        parameters = {}

    from pm4py.objects.bpmn.obj import BPMN
    from pm4py.objects.bpmn.util.sorting import get_sorted_nodes_edges

    image_format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    rankdir = exec_utils.get_param_value(Parameters.RANKDIR, parameters, constants.DEFAULT_RANKDIR_GVIZ)
    font_size = exec_utils.get_param_value(Parameters.FONT_SIZE, parameters, 28)
    font_size = str(font_size)
    bgcolor = exec_utils.get_param_value(Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR)

    filename = tempfile.NamedTemporaryFile(suffix='.gv')
    filename.close()

    viz = Digraph("", filename=filename.name, engine='dot', graph_attr={'bgcolor': bgcolor})
    viz.graph_attr['rankdir'] = rankdir
    # viz.attr(nodesep='1')
    # viz.attr(ranksep='0.4')

    nodes, edges = get_sorted_nodes_edges(bpmn_graph)

    for n in nodes:
        n_id = str(id(n))
        if isinstance(n, BPMN.Task):
            viz.node(n_id, shape="box", label=get_label(n.get_name()), fontsize=font_size)
        elif isinstance(n, BPMN.StartEvent):
            viz.node(n_id, label="", shape="circle", fontsize=font_size, width='0.6', height='0.6')
        elif isinstance(n, BPMN.EndEvent):
            viz.node(n_id, label="", shape="circle", fontsize=font_size, penwidth="3.0", width='0.6', height='0.6')
        elif isinstance(n, BPMN.ParallelGateway):
            with importlib.resources.path("powl.visualization.powl.variants.icons", "gate_and.svg") as gimg:
                xor_image = str(gimg)
                viz.node(n_id, label="", shape="diamond",
                         width='0.8', height='0.8', fixedsize="true", image=xor_image)
            # viz.node(n_id, label="+", shape="diamond", fontsize=font_size)
        elif isinstance(n, BPMN.ExclusiveGateway):
            with importlib.resources.path("powl.visualization.powl.variants.icons", "gate.svg") as gimg:
                xor_image = str(gimg)
                viz.node(n_id, label="", shape="diamond",
                         width='0.8', height='0.8', fixedsize="true", image=xor_image)
            # viz.node(n_id, label="X", shape="diamond", fontsize=font_size)
        elif isinstance(n, BPMN.InclusiveGateway):
            viz.node(n_id, label="O", shape="diamond", fontsize=font_size)
        else:
            viz.node(n_id, label="", shape="circle", fontsize=font_size)

    for e in edges:
        n_id_1 = str(id(e[0]))
        n_id_2 = str(id(e[1]))

        viz.edge(n_id_1, n_id_2, penwidth="2.0")

    viz.attr(overlap='false')

    viz.format = image_format.replace("html", "plain-ext")

    return viz


def save(gviz: graphviz.Digraph, output_file_path: str, parameters=None):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path, parameters=parameters)
    return ""


def view(gviz: graphviz.Digraph, parameters=None):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    if True:
        return gview.view(gviz, parameters=parameters)


def matplotlib_view(gviz: graphviz.Digraph, parameters=None):
    """
    Views the diagram using Matplotlib

    Parameters
    ---------------
    gviz
        Graphviz
    """
    if True:
        return gview.matplotlib_view(gviz, parameters=parameters)



from abc import ABC
from typing import Any, Optional, Dict, Generic

from pm4py.algo.discovery.inductive.cuts.concurrency import ConcurrencyCut, ConcurrencyCutUVCL, T
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from powl.objects.obj import StrictPartialOrder


class POWLConcurrencyCut(ConcurrencyCut, ABC, Generic[T]):

    @classmethod
    def operator(cls, parameters: Optional[Dict[str, Any]] = None) -> StrictPartialOrder:
        return StrictPartialOrder([])


class POWLConcurrencyCutUVCL(ConcurrencyCutUVCL, POWLConcurrencyCut[IMDataStructureUVCL]):
    pass




from collections import defaultdict

from powl.discovery.partial_order_based.utils.constants import VARIANT_FREQUENCY_KEY
from powl.discovery.partial_order_based.utils.simplified_objects import Graph, Skip


class NodeGrouping:

    @classmethod
    def find_groups(cls, partial_orders):

        partial_orders = list(partial_orders)
        node_to_orders = defaultdict(list)
        n = len(partial_orders)

        all_nodes = list({node for graph in partial_orders for node in graph.nodes})


        for node_id, current_node in enumerate(all_nodes):
            for graph_id, graph in enumerate(partial_orders):
                if current_node in graph.nodes:
                    node_to_orders[node_id].append(graph_id)

        graph_ids_lists_to_nodes = defaultdict(list)

        sorted_keys = sorted(node_to_orders.keys(), key=lambda x: len(node_to_orders[x]), reverse=True)

        for node_id in sorted_keys:
            graph_id_list = node_to_orders[node_id]


            new_frozenset = frozenset(graph_id_list)

            graph_ids_lists_to_nodes[new_frozenset].append(node_id)

        res_dict = {}
        new_nodes_counter = defaultdict(int)
        if len(graph_ids_lists_to_nodes.keys()) == 1:
            pass
        else:
            for graph_id_list, node_id_list in graph_ids_lists_to_nodes.items():
                all_projections = []
                for graph_id in graph_id_list:
                    graph = partial_orders[graph_id]
                    proj_nodes = [node for i, node in enumerate(all_nodes) if i in node_id_list and node in graph.nodes]
                    proj_edges = [(s,t) for (s, t) in graph.edges if s in proj_nodes and t in proj_nodes]
                    projection = Graph(frozenset(proj_nodes),
                                       frozenset(proj_edges),
                                       {VARIANT_FREQUENCY_KEY: graph.additional_information[VARIANT_FREQUENCY_KEY]})
                    all_projections.append(projection)
                from powl.discovery.partial_order_based.variants.base.miner import _mine
                new_graph = _mine(all_projections)

                if len(graph_id_list) < n:
                    xor = Skip.create(new_graph)
                    new_node = xor
                else:
                    new_node = new_graph

                for node_id in node_id_list:
                    node = all_nodes[node_id]
                    res_dict[node] = new_node
                new_nodes_counter[new_node] += 1

        for node in all_nodes:
            if node not in res_dict.keys():
                from powl.discovery.partial_order_based.variants.base.miner import apply_mining_algorithm_recursively
                new_node = apply_mining_algorithm_recursively(node)
                res_dict[node] = new_node
                new_nodes_counter[new_node] += 1

        return res_dict, new_nodes_counter



from powl.conversion.variants import to_petri_net
from pm4py.util import exec_utils
from enum import Enum


class Variants(Enum):
    TO_PETRI_NET = to_petri_net


def apply(powl, parameters=None, variant=Variants.TO_PETRI_NET):
    """
    Method for converting from POWL to Petri net

    Parameters
    -----------
    powl
        POWL model
    parameters
        Parameters of the algorithm
    variant
        Chosen variant of the algorithm:
            - Variants.TO_PETRI_NET

    Returns
    -----------
    net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    """
    return exec_utils.get_variant(variant).apply(powl, parameters=parameters)



from typing import List, Optional, Dict, Any, Tuple

from powl.discovery.total_order_based.inductive.cuts.factory import T, CutFactory
from powl.discovery.total_order_based.inductive.cuts.loop import POWLLoopCutUVCL
from powl.discovery.total_order_based.inductive.cuts.xor import POWLExclusiveChoiceCutUVCL
from powl.discovery.total_order_based.inductive.variants.dynamic_clustering.dynamic_clustering_partial_order_cut import \
    DynamicClusteringPartialOrderCutUVCL
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructure
from powl.objects.obj import POWL
from pm4py.objects.dfg import util as dfu


class CutFactoryPOWLDynamicClustering(CutFactory):

    @classmethod
    def get_cuts(cls, obj, parameters=None):
        return [POWLExclusiveChoiceCutUVCL, POWLLoopCutUVCL, DynamicClusteringPartialOrderCutUVCL]

    @classmethod
    def find_cut(cls, obj: IMDataStructure, parameters: Optional[Dict[str, Any]] = None) -> Optional[
            Tuple[POWL, List[T]]]:
        alphabet = sorted(dfu.get_vertices(obj.dfg), key=lambda g: g.__str__())
        if len(alphabet) < 2:
            return None
        for c in CutFactoryPOWLDynamicClustering.get_cuts(obj):
            r = c.apply(obj, parameters)
            if r is not None:
                return r
        return None



from itertools import combinations
from typing import Union

from pm4py.objects.petri_net.obj import PetriNet
from pm4py.objects.petri_net.utils import petri_utils as pn_util


def id_generator():
    count = 1
    while True:
        yield f"id{count}"
        count += 1


def add_arc_from_to(source: Union[PetriNet.Place, PetriNet.Transition],
                    target: Union[PetriNet.Transition, PetriNet.Place], net: PetriNet):
    arc = PetriNet.Arc(source, target)
    net.arcs.add(arc)
    source.out_arcs.add(arc)
    target.in_arcs.add(arc)


def get_pre_places(place):
    preset_transitions = pn_util.pre_set(place)
    preset_places = set()
    for t in preset_transitions:
        if t.label or len(t.in_arcs) != 1 or len(t.out_arcs) != 1:
            return None, None
        else:
            preset_places.add(list(t.in_arcs)[0].source)
    return preset_places, preset_transitions


def get_post_nodes(place):
    postset_transitions = pn_util.post_set(place)
    postset_places = set()
    for t in postset_transitions:
        if t.label or len(t.in_arcs) != 1 or len(t.out_arcs) != 1:
            return None, None
        else:
            postset_places.add(list(t.out_arcs)[0].target)
    return postset_places, postset_transitions


def merge_places_connected_with_silent_transition(net, source, sink):
    all_silent_transitions = [t for t in net.transitions if not t.label]

    reduced = True
    while reduced:
        reduced = False
        for t in all_silent_transitions:
            in_arcs = list(t.in_arcs)
            out_arcs = list(t.out_arcs)

            if len(in_arcs) == len(out_arcs) == 1:
                p1 = in_arcs[0].source
                p2 = out_arcs[0].target

                if p1 != source and p2 != sink:
                    if len(p1.out_arcs) == 1:
                        for arc in list(p1.in_arcs):
                            add_arc_from_to(arc.source, p2, net)
                        pn_util.remove_place(net, p1)
                        pn_util.remove_transition(net, t)
                        reduced = True
                        break
                    elif len(p2.in_arcs) == 1:
                        for arc in list(p2.out_arcs):
                            add_arc_from_to(p1, arc.target, net)
                        pn_util.remove_place(net, p2)
                        pn_util.remove_transition(net, t)
                        reduced = True
                        break
    return net


def merge_places_with_identical_preset_or_postset(net, source, sink):
    all_places = [p for p in net.places if p not in {source, sink}]
    for p1, p2 in combinations(all_places, 2):
        pre1_p, pre1_t = get_pre_places(p1)
        pre2_p, pre2_t = get_pre_places(p2)
        post1_p, post1_t = get_post_nodes(p1)
        post2_p, post2_t = get_post_nodes(p2)

        if pre1_p and pre2_p and pre1_p == pre2_p:
            for t in pn_util.post_set(p2):
                add_arc_from_to(p1, t, net)
            for silent_t in pre2_t:
                pn_util.remove_transition(net, silent_t)
            pn_util.remove_place(net, p2)
            return merge_places_with_identical_preset_or_postset(net, source, sink)

        if post1_p and post2_p and post1_p == post2_p:
            for t in pn_util.pre_set(p2):
                add_arc_from_to(t, p1, net)
            for silent_t in post2_t:
                pn_util.remove_transition(net, silent_t)
            pn_util.remove_place(net, p2)
            return merge_places_with_identical_preset_or_postset(net, source, sink)

    return net



import warnings
import pandas as pd
import pm4py
from pm4py.algo.discovery.inductive.variants.imf import IMFParameters
from powl.discovery.total_order_based.inductive.utils.filtering import FILTERING_THRESHOLD
from powl.discovery.total_order_based.inductive.variants.dynamic_clustering_frequency.dynamic_clustering_frequency_partial_order_cut import \
    ORDER_FREQUENCY_RATIO
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import POWLDiscoveryVariant
from powl.objects.obj import POWL
from pm4py.objects.ocel.obj import OCEL
from pm4py.utils import get_properties
from powl.conversion.converter import apply as powl_converter
from powl.visualization.powl.visualizer import POWLVisualizationVariants
from pm4py.objects.bpmn.layout import layouter
from powl.conversion.variants.to_bpmn import apply as bpmn_converter
from powl.discovery.object_centric.algorithm import apply as oc_discovery


def import_ocel(path: str) -> OCEL:
    try:
        ocel = pm4py.read_ocel2(path)
    except Exception as e:
        ocel = pm4py.read_ocel(path)
    return ocel


def import_event_log(path: str, timestamp_key = None) -> pd.DataFrame:
    import rustxes
    if path.endswith(".xes") or path.endswith(".xes.gz"):
        [xes, log_attrs] = rustxes.import_xes(path)
        df = xes.to_pandas()
    elif path.endswith(".csv"):
        cols_to_parse = []
        df_sample = pd.read_csv(path, nrows=0)

        if timestamp_key:
            if timestamp_key not in df_sample.columns:
                raise ValueError("Timestamp key not found in table!")
        elif 'time:timestamp' in df_sample.columns:
            timestamp_key =  'time:timestamp'
            warnings.warn(f"No column given as a timestamp key! Default value 'time:timestamp' is used.")
        else:
            warnings.warn("No column given as a timestamp key! Timestamp column parsing is skipped!")

        if timestamp_key:
            cols_to_parse = [timestamp_key]

        df = pd.read_csv(path, keep_default_na=False, parse_dates=cols_to_parse)
    else:
        raise ValueError("Unsupported file type!")
    return df


def discover(log: pd.DataFrame, variant=POWLDiscoveryVariant.DECISION_GRAPH_MAX,
                  filtering_weight_factor: float = None, order_graph_filtering_threshold: float = None,
                  dfg_frequency_filtering_threshold: float = None,
                  activity_key: str = "concept:name", timestamp_key: str = "time:timestamp",
                  case_id_key: str = "case:concept:name",
                  lifecycle_key: str = "lifecycle:transition",
                  keep_only_completion_events: bool = True,
                  ) -> POWL:
    """
    Discovers a POWL model from an event log.

    Reference paper:
    H Kourani, G Park, WMP van der Aalst. "Unlocking Non-Block-Structured Decisions: Inductive Mining with Choice Graphs" arXiv preprint arXiv:2505.07052.

    :param keep_only_completion_events:
    :param lifecycle_key:
    :param log: event log / Pandas dataframe
    :param variant: variant of the algorithm
    :param filtering_weight_factor: accepts values 0 <= x < 1
    :param order_graph_filtering_threshold: accepts values 0.5 < x <= 1
    :param dfg_frequency_filtering_threshold: accepts values 0 <= x < 1
    :param activity_key: attribute to be used for the activity
    :param timestamp_key: attribute to be used for the timestamp
    :param case_id_key: attribute to be used as case identifier
    :rtype: ``POWL``
    """

    log = log.sort_values([case_id_key, timestamp_key])
    properties = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key)

    if keep_only_completion_events and lifecycle_key in log.columns:
        filtered_log = log[log["lifecycle:transition"].isin(["complete", "COMPLETE", "Complete"])]
        if len(filtered_log) > 0:
            log = filtered_log

    from powl.discovery.total_order_based.inductive.utils.filtering import FILTERING_TYPE, FilteringType

    num_filters = 0
    if order_graph_filtering_threshold is not None:
        if not variant is POWLDiscoveryVariant.DYNAMIC_CLUSTERING:
            raise Exception("The order graph filtering threshold can only be used for the variant DYNAMIC_CLUSTERING!")
        properties[ORDER_FREQUENCY_RATIO] = order_graph_filtering_threshold
        properties[FILTERING_TYPE] = FilteringType.DYNAMIC
        num_filters += 1
    if dfg_frequency_filtering_threshold is not None:
        properties[IMFParameters.NOISE_THRESHOLD] = dfg_frequency_filtering_threshold
        properties[FILTERING_TYPE] = FilteringType.DFG_FREQUENCY
        num_filters += 1
    if filtering_weight_factor is not None:
        properties[FILTERING_THRESHOLD] = filtering_weight_factor
        properties[FILTERING_TYPE] = FilteringType.DECREASING_FACTOR
        num_filters += 1

    if num_filters > 1:
        raise Exception("The algorithm can only be used with one filtering threshold at a time!")

    from powl.discovery.total_order_based import algorithm as powl_discovery
    return powl_discovery.apply(log, variant=variant, parameters=properties)


def discover_petri_net_from_ocel(ocel: OCEL, parameters):
    return oc_discovery(ocel, parameters=parameters)


def discover_from_partially_ordered_log(log: pd.DataFrame,
      activity_key: str = "concept:name",
      order_key: str = "time:timestamp",
      case_id_key: str = "case:concept:name",
      lifecycle_key: str|None = "lifecycle:transition"
      ) -> POWL:
    """
    Discovers a POWL model from a partially ordered event log.

    Reference paper:
    H Kourani, G Park, WMP van der Aalst. "Revealing Inherent Concurrency in Event Data: A Partial Order Approach to Process Discovery"

    :param log: event log / Pandas dataframe
    :param activity_key: attribute to be used for the activity
    :param order_key: attribute to be used for ordering events within traces
    :param case_id_key: attribute to be used as case identifier
    :param lifecycle_key: attribute to be used as lifecycle identifier
    :rtype: ``POWL``
    """
    complete_tags = {"complete", "COMPLETE", "Complete"}
    start_tags = {"start", "START", "Start"}

    from powl.discovery.partial_order_based.utils import log_to_partial_orders
    partial_orders = log_to_partial_orders.apply(log,
                                                 case_id_col=case_id_key,
                                                 activity_col=activity_key,
                                                 ordering_col=order_key,
                                                 lifecycle_col=lifecycle_key,
                                                 start_transitions=start_tags,
                                                 complete_transitions=complete_tags)
    from powl.discovery.partial_order_based.variants.base import miner
    powl = miner.apply(partial_orders)
    return powl


def view(powl: POWL, use_frequency_tags=True):
    from powl.visualization.powl import visualizer as powl_visualizer
    gviz = powl_visualizer.apply(powl, variant=POWLVisualizationVariants.BASIC, frequency_tags=use_frequency_tags)
    powl_visualizer.view(gviz)


def view_net(powl: POWL, use_frequency_tags=True):
    from powl.visualization.powl import visualizer as powl_visualizer
    gviz = powl_visualizer.apply(powl, variant=POWLVisualizationVariants.NET, frequency_tags=use_frequency_tags)
    powl_visualizer.view(gviz)


def save_visualization(powl: POWL, file_path: str, use_frequency_tags=True):
    file_path = str(file_path)
    from powl.visualization.powl import visualizer as powl_visualizer
    gviz = powl_visualizer.apply(powl, variant=POWLVisualizationVariants.BASIC, frequency_tags=use_frequency_tags)
    return powl_visualizer.save(gviz, file_path)


def save_visualization_net(powl: POWL, file_path: str, use_frequency_tags=True):
    file_path = str(file_path)
    from powl.visualization.powl import visualizer as powl_visualizer
    gviz = powl_visualizer.apply(powl, variant=POWLVisualizationVariants.NET, frequency_tags=use_frequency_tags)
    return powl_visualizer.save(gviz, file_path)


def convert_to_petri_net(powl: POWL):
    return powl_converter(powl)


def convert_to_bpmn(powl: POWL):
    bpmn, _, _ = bpmn_converter(powl)
    bpmn = layouter.apply(bpmn)
    return bpmn



from typing import Optional, Tuple, List, Dict, Any, Type

from pm4py.algo.discovery.inductive.fall_through.empty_traces import EmptyTracesUVCL
from powl.discovery.total_order_based.inductive.fall_through.decision_graph.empty_traces_decision_graph import \
    POWLEmptyTracesDecisionGraphUVCL
from powl.discovery.total_order_based.inductive.variants.decision_graph.factory_dg_max import CutFactoryPOWLDecisionGraphMaximal
from powl.discovery.total_order_based.inductive.variants.im_tree import IMBasePOWL, T
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import POWLDiscoveryVariant
from powl.objects.obj import POWL


class POWLInductiveMinerDecisionGraphMaximal(IMBasePOWL):

    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.DECISION_GRAPH_MAX

    def empty_traces_cut(self) -> Type[EmptyTracesUVCL]:
        return POWLEmptyTracesDecisionGraphUVCL

    def find_cut(self, obj: T, parameters: Optional[Dict[str, Any]] = None) -> Optional[Tuple[POWL, List[T]]]:
        res = CutFactoryPOWLDecisionGraphMaximal.find_cut(obj, parameters=parameters)
        return res




from abc import ABC
from typing import Any, Optional, Dict, Tuple, List, Generic

from pm4py.algo.discovery.inductive.base_case.abc import T
from pm4py.algo.discovery.inductive.cuts.sequence import SequenceCut, SequenceCutUVCL, StrictSequenceCutUVCL, \
    StrictSequenceCut
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from powl.objects.obj import Sequence


class POWLSequenceCut(SequenceCut, ABC, Generic[T]):

    @classmethod
    def operator(cls, parameters: Optional[Dict[str, Any]] = None) -> Sequence:
        return Sequence([])

    @classmethod
    def apply(cls, obj: T, parameters: Optional[Dict[str, Any]] = None) -> Optional[Tuple[Sequence, List[T]]]:
        g = cls.holds(obj, parameters)
        if g is None:
            return g
        children = cls.project(obj, g, parameters)
        po = Sequence(children)
        return po, children


class POWLStrictSequenceCut(POWLSequenceCut[T], StrictSequenceCut, ABC):
    pass


class POWLSequenceCutUVCL(SequenceCutUVCL, POWLSequenceCut[IMDataStructureUVCL]):
    pass


class POWLStrictSequenceCutUVCL(StrictSequenceCutUVCL, StrictSequenceCut[IMDataStructureUVCL], POWLSequenceCutUVCL):
    pass



from multiprocessing import Pool, Manager
from typing import List, TypeVar, Tuple, Optional, Dict, Any, Type

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructure, IMDataStructureUVCL
from pm4py.algo.discovery.inductive.fall_through.abc import FallThrough
from powl.discovery.total_order_based.inductive.fall_through.activity_concurrent import POWLActivityConcurrentUVCL
from powl.discovery.total_order_based.inductive.fall_through.activity_once_per_trace import POWLActivityOncePerTraceUVCL
from powl.discovery.total_order_based.inductive.fall_through.flower import POWLFlowerModelUVCL
from powl.discovery.total_order_based.inductive.fall_through.strict_tau_loop import POWLStrictTauLoopUVCL
from powl.discovery.total_order_based.inductive.fall_through.tau_loop import POWLTauLoopUVCL
from powl.objects.obj import POWL

T = TypeVar('T', bound=IMDataStructure)
S = TypeVar('S', bound=FallThrough)


class FallThroughFactory:

    @classmethod
    def get_fall_throughs(cls, obj: T, parameters: Optional[Dict[str, Any]] = None) -> List[Type[S]]:
        if type(obj) is IMDataStructureUVCL:
            return [POWLActivityOncePerTraceUVCL, POWLActivityConcurrentUVCL,
                    POWLStrictTauLoopUVCL, POWLTauLoopUVCL, POWLFlowerModelUVCL]
        return list()

    @classmethod
    def fall_through(cls, obj: T, pool: Pool, manager: Manager, parameters: Optional[Dict[str, Any]] = None) -> Optional[Tuple[POWL, List[T]]]:
        for f in FallThroughFactory.get_fall_throughs(obj):
            r = f.apply(obj, pool, manager, parameters)
            if r is not None:
                return r
        return None



from multiprocessing import Pool, Manager
from typing import Tuple, List, Optional, Dict, Any

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.algo.discovery.inductive.fall_through.empty_traces import EmptyTracesUVCL
from powl.objects.BinaryRelation import BinaryRelation
from powl.objects.obj import DecisionGraph
from copy import copy


class POWLEmptyTracesDecisionGraphUVCL(EmptyTracesUVCL):

    @classmethod
    def apply(cls, obj: IMDataStructureUVCL, pool: Pool = None, manager: Manager = None,
              parameters: Optional[Dict[str, Any]] = None) -> Optional[
        Tuple[DecisionGraph, List[IMDataStructureUVCL]]]:
        if cls.holds(obj, parameters):
            data_structure = copy(obj.data_structure)
            del data_structure[()]
            children = [IMDataStructureUVCL(data_structure)]
            dg = DecisionGraph(BinaryRelation(copy(children)), children, children, empty_path=True)
            return dg, children
        else:
            return None



from itertools import product
from abc import ABC
from itertools import combinations
from typing import Any, Optional, Dict, List, Generic, Tuple, Collection

from pm4py.algo.discovery.inductive.cuts.abc import Cut, T
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from powl.objects.BinaryRelation import BinaryRelation
from powl.objects.obj import StrictPartialOrder, POWL
from pm4py.algo.discovery.inductive.cuts import utils as cut_util
from powl.discovery.total_order_based.inductive.variants.maximal.maximal_partial_order_cut import project_on_groups_with_unique_activities
from pm4py.objects.dfg import util as dfu
from pm4py.statistics.eventually_follows.uvcl.get import apply as to_efg


def generate_order(clusters, efg):
    # Step 0: if we have one single group containing all activities ---> invoke fall-through.
    if len(clusters) < 2:
        return None

    # Step 1: Generate Order based on the EFG
    po = BinaryRelation([tuple(c) for c in sorted(clusters)])
    at_least_one_efg = [[False for _ in range(len(po.nodes))] for _ in range(len(po.nodes))]
    all_efg = [[True for _ in range(len(po.nodes))] for _ in range(len(po.nodes))]

    changed = False
    for i in range(len(po.nodes)):
        cluster_1 = po.nodes[i]
        for j in range(i + 1, len(po.nodes)):
            cluster_2 = po.nodes[j]
            for a in cluster_1:
                for b in cluster_2:
                    if (a, b) in efg:
                        at_least_one_efg[i][j] = True
                    else:
                        all_efg[i][j] = False
                    if (b, a) in efg:
                        at_least_one_efg[j][i] = True
                    else:
                        all_efg[j][i] = False

            if at_least_one_efg[i][j] and not at_least_one_efg[j][i]:
                po.add_edge(cluster_1, cluster_2)
            elif at_least_one_efg[j][i] and not at_least_one_efg[i][j]:
                po.add_edge(cluster_2, cluster_1)

    # Step 2: Ensure Transitivity and Irreflexivity
    if not po.is_transitive():
        n = len(po.nodes)
        continue_loop = True
        while continue_loop:
            continue_loop = False
            for i, j, k in product(range(n), range(n), range(n)):
                if i != j and j != k and po.edges[i][j] and po.edges[j][k] and not po.is_edge_id(i, k):
                    if not at_least_one_efg[k][i]:
                        po.edges[i][k] = True
                        continue_loop = True
                    else:
                        clusters = cut_util.merge_lists_based_on_activities(po.nodes[i][0], po.nodes[k][0], clusters)
                        return generate_order(clusters, efg)

    if not po.is_irreflexive():
        for i in range(len(po.nodes)):
            cluster_1 = po.nodes[i]
            for j in range(i + 1, len(po.nodes)):
                cluster_2 = po.nodes[j]
                if po.is_edge(cluster_1, cluster_2) and po.is_edge(cluster_2, cluster_1):
                    clusters = cut_util.merge_lists_based_on_activities(cluster_1[0], cluster_2[0], clusters)
                    changed = True

    if changed:
        return generate_order(clusters, efg)

    # # Step 3: Detect Choice
    for i in range(len(po.nodes)):
        cluster_1 = po.nodes[i]
        for j in range(i + 1, len(po.nodes)):
            cluster_2 = po.nodes[j]
            if not po.is_edge(cluster_1, cluster_2) and not po.is_edge(cluster_2, cluster_1) and not \
                    at_least_one_efg[i][j] and not at_least_one_efg[j][i]:
                clusters = cut_util.merge_lists_based_on_activities(cluster_1[0], cluster_2[0], clusters)
                changed = True

    if changed:
        return generate_order(clusters, efg)

    # Step 4: Cluster nodes sharing the same pre- and post-sets.
    pre = {node: [] for node in po.nodes}
    post = {node: [] for node in po.nodes}
    for i in range(len(po.nodes)):
        cluster_1 = po.nodes[i]
        for j in range(i + 1, len(po.nodes)):
            cluster_2 = po.nodes[j]
            if po.is_edge(cluster_1, cluster_2):
                pre[cluster_2].append(cluster_1)
                post[cluster_1].append(cluster_2)
            elif po.is_edge(cluster_2, cluster_1):
                pre[cluster_1].append(cluster_2)
                post[cluster_2].append(cluster_1)

    for i in range(len(po.nodes)):
        cluster_1 = po.nodes[i]
        for j in range(i + 1, len(po.nodes)):
            cluster_2 = po.nodes[j]
            if pre[cluster_1] == pre[cluster_2] and post[cluster_1] == post[cluster_2]:
                clusters = cut_util.merge_lists_based_on_activities(cluster_1[0], cluster_2[0], clusters)
                changed = True

    if changed and len(clusters) > 1:
        return generate_order(clusters, efg)
    else:
        return po


class DynamicClusteringPartialOrderCut(Cut[T], ABC, Generic[T]):

    @classmethod
    def operator(cls, parameters: Optional[Dict[str, Any]] = None) -> StrictPartialOrder:
        return StrictPartialOrder([])

    @classmethod
    def holds(cls, obj: T, parameters: Optional[Dict[str, Any]] = None) -> Optional[BinaryRelation]:
        alphabet = sorted(dfu.get_vertices(obj.dfg), key=lambda g: g.__str__())
        efg = to_efg(obj)
        clusters = [[a] for a in alphabet]
        po = generate_order(clusters, efg)
        return po

    @classmethod
    def apply(cls, obj: T, parameters: Optional[Dict[str, Any]] = None) -> Optional[Tuple[StrictPartialOrder,
                                                                                          List[POWL]]]:
        g = cls.holds(obj, parameters)
        if g is None:
            return g
        children = cls.project(obj, g.nodes, parameters)
        po = StrictPartialOrder(children)
        for i, j in combinations(range(len(g.nodes)), 2):
            if g.is_edge_id(i, j):
                po.order.add_edge(children[i], children[j])
            elif g.is_edge_id(j, i):
                po.order.add_edge(children[j], children[i])
        return po, po.children


class DynamicClusteringPartialOrderCutUVCL(DynamicClusteringPartialOrderCut[IMDataStructureUVCL]):

    @classmethod
    def project(cls, obj: IMDataStructureUVCL, groups: List[Collection[Any]],
                parameters: Optional[Dict[str, Any]] = None) -> List[IMDataStructureUVCL]:
        return project_on_groups_with_unique_activities(obj.data_structure, groups)



import os
from setuptools import setup, find_packages

# Read the requirements from the requirements.txt file
with open('requirements.txt', encoding='utf-8') as f:
    requirements = f.read().splitlines()

# Read the long description from README.md
with open('README.md', encoding='utf-8') as f:
    long_description = f.read()


# Get the version from __init__.py
def get_version():
    with open(os.path.join('powl', '__init__.py'), 'r') as f:
        for line in f:
            if line.startswith('__version__'):
                _, version = line.split("=")
                return version.strip().replace("'", "").replace('"', '')
    raise RuntimeError("Unable to find __version__ string.")


version = get_version()

setup(
    name='powl',
    version=version,
    description='POWL Miner: Process Mining with the Partially Ordered Workflow Language',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Humam Kourani',
    author_email='humam.kourani@gmail.com',
    url='https://github.com/humam-kourani/powl',
    license='AGPL-3.0 license',
    packages=find_packages(),
    install_requires=requirements,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
    keywords='process mining, business process management',
    python_requires='>=3.9',
    # Include any data files (if any)
    # For example: package_data={'your_package': ['data/*.txt']},
    # Or: include_package_data=True  (to include everything in MANIFEST.in)
    # Entry points for command-line scripts (if any)
    # entry_points={
    #     'console_scripts': [
    #         'your_script_name = your_package.module:main_function',
    #     ],
    # },
)


from enum import Enum
from pm4py.objects.ocel.obj import OCEL
from pm4py.util import exec_utils
from typing import Optional, Dict, Any
from powl.discovery.object_centric.variants.flattening import miner as flatten_miner


class Variants(Enum):
    FLATTENING = flatten_miner


def apply(
    ocel: OCEL,
    variant=Variants.FLATTENING,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, Any]:
    """
    Discovers an object-centric Petri net from the provided object-centric event log.

    Parameters
    -----------------
    ocel
        Object-centric event log
    variant
        Variant of the algorithm to be used
    parameters
        Variant-specific parameters

    Returns
    ----------------
    ocpn
        Object-centric Petri net model, as a dictionary of properties.
    """
    return exec_utils.get_variant(variant).apply(ocel, parameters=parameters)


from typing import Optional, Tuple, List, TypeVar, Dict, Any

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureLog
from powl.discovery.total_order_based.inductive.variants.dynamic_clustering.factory import CutFactoryPOWLDynamicClustering
from powl.discovery.total_order_based.inductive.variants.im_tree import IMBasePOWL
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import POWLDiscoveryVariant
from powl.objects.obj import POWL

T = TypeVar('T', bound=IMDataStructureLog)


class POWLInductiveMinerDynamicClustering(IMBasePOWL):

    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.DYNAMIC_CLUSTERING

    def find_cut(self, obj: T, parameters: Optional[Dict[str, Any]] = None) -> Optional[Tuple[POWL, List[T]]]:
        res = CutFactoryPOWLDynamicClustering.find_cut(obj, parameters=parameters)
        return res



import tempfile
from copy import copy

from graphviz import Digraph

from pm4py.statistics.attributes.log import get as attr_get
from pm4py.objects.dfg.utils import dfg_utils
from pm4py.util import xes_constants as xes
from pm4py.util import exec_utils
from pm4py.statistics.service_time.log import get as serv_time_get
from enum import Enum
from pm4py.util import constants
from typing import Optional, Dict, Any, Tuple
import graphviz
from pm4py.objects.log.obj import EventLog
from collections import Counter


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    FORMAT = "format"
    MAX_NO_EDGES_IN_DIAGRAM = "maxNoOfEdgesInDiagram"
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    FONT_SIZE = "font_size"
    RANKDIR = "rankdir"
    BGCOLOR = "bgcolor"


def apply(dfg: Dict[Tuple[str, str], int], log: EventLog = None, parameters: Optional[Dict[Any, Any]] = None, activities_count : Dict[str, int] = None, serv_time: Dict[str, float] = None) -> graphviz.Digraph:
    """
    Visualize a frequency directly-follows graph

    Parameters
    -----------------
    dfg
        Frequency Directly-follows graph
    log
        (if provided) Event log for the calculation of statistics
    activities_count
        (if provided) Dictionary associating to each activity the number of occurrences in the log.
    serv_time
        (if provided) Dictionary associating to each activity the average service time
    parameters
        Variant-specific parameters

    Returns
    -----------------
    gviz
        Graphviz digraph
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY)
    image_format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    max_no_of_edges_in_diagram = exec_utils.get_param_value(Parameters.MAX_NO_EDGES_IN_DIAGRAM, parameters, 100000)
    start_activities = exec_utils.get_param_value(Parameters.START_ACTIVITIES, parameters, {})
    end_activities = exec_utils.get_param_value(Parameters.END_ACTIVITIES, parameters, {})
    font_size = exec_utils.get_param_value(Parameters.FONT_SIZE, parameters, 32)
    font_size = str(font_size)

    if start_activities is None:
        start_activities = dict()
    if end_activities is None:
        end_activities = dict()
    activities = sorted(list(set(dfg_utils.get_activities_from_dfg(dfg)).union(set(start_activities)).union(set(end_activities))))

    rankdir = exec_utils.get_param_value(Parameters.RANKDIR, parameters, constants.DEFAULT_RANKDIR_GVIZ)
    bgcolor = exec_utils.get_param_value(Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR)

    if activities_count is None:
        if log is not None:
            activities_count = attr_get.get_attribute_values(log, activity_key, parameters=parameters)
        else:
            # the frequency of an activity in the log is at least the number of occurrences of
            # incoming arcs in the DFG.
            # if the frequency of the start activities nodes is also provided, use also that.
            activities_count = Counter({key: 0 for key in activities})
            for el in dfg:
                activities_count[el[1]] += dfg[el]
            if isinstance(start_activities, dict):
                for act in start_activities:
                    activities_count[act] += start_activities[act]

    if serv_time is None:
        if log is not None:
            serv_time = serv_time_get.apply(log, parameters=parameters)
        else:
            serv_time = {key: 0 for key in activities}

    return graphviz_visualization(activities_count, dfg, image_format=image_format, measure="frequency",
                                           max_no_of_edges_in_diagram=max_no_of_edges_in_diagram,
                                           start_activities=start_activities, end_activities=end_activities, serv_time=serv_time,
                                           font_size=font_size, bgcolor=bgcolor, rankdir=rankdir)


def graphviz_visualization(activities_count, dfg, image_format="png", measure="frequency",
                           max_no_of_edges_in_diagram=100000, start_activities=None, end_activities=None, serv_time=None,
                           font_size="12", bgcolor=constants.DEFAULT_BGCOLOR, rankdir=constants.DEFAULT_RANKDIR_GVIZ):
    """
    Do GraphViz visualization of a DFG graph

    Parameters
    -----------
    activities_count
        Count of attributes in the log (may include attributes that are not in the DFG graph)
    dfg
        DFG graph
    image_format
        GraphViz should be represented in this format
    measure
        Describes which measure is assigned to edges in direcly follows graph (frequency/performance)
    max_no_of_edges_in_diagram
        Maximum number of edges in the diagram allowed for visualization
    start_activities
        Start activities of the log
    end_activities
        End activities of the log
    serv_time
        For each activity, the service time in the log
    font_size
        Size of the text on the activities/edges
    bgcolor
        Background color of the visualization (i.e., 'transparent', 'white', ...)
    rankdir
        Direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)

    Returns
    -----------
    viz
        Digraph object
    """
    if start_activities is None:
        start_activities = []
    if end_activities is None:
        end_activities = []

    filename = tempfile.NamedTemporaryFile(suffix='.gv')
    filename.close()

    viz = Digraph("", filename=filename.name, engine='dot', graph_attr={'bgcolor': bgcolor, 'rankdir': rankdir})

    # first, remove edges in diagram that exceeds the maximum number of edges in the diagram
    dfg_key_value_list = []
    for edge in dfg:
        dfg_key_value_list.append([edge, dfg[edge]])
    # more fine grained sorting to avoid that edges that are below the threshold are
    # undeterministically removed
    dfg_key_value_list = sorted(dfg_key_value_list, key=lambda x: (x[1], x[0][0], x[0][1]), reverse=True)
    dfg_key_value_list = dfg_key_value_list[0:min(len(dfg_key_value_list), max_no_of_edges_in_diagram)]
    dfg_allowed_keys = [x[0] for x in dfg_key_value_list]
    dfg_keys = list(dfg.keys())
    for edge in dfg_keys:
        if edge not in dfg_allowed_keys:
            del dfg[edge]


    activities_count_int = copy(activities_count)

    activities_in_dfg = set(activities_count)

    # represent nodes
    viz.attr('node', shape='box')

    if len(activities_in_dfg) == 0:
        activities_to_include = sorted(list(set(activities_count_int)))
    else:
        # take unique elements as a list not as a set (in this way, nodes are added in the same order to the graph)
        activities_to_include = sorted(list(set(activities_in_dfg)))

    activities_map = {}

    for act in activities_to_include:
        viz.node(str(hash(act)), act, fontsize=font_size)
        activities_map[act] = str(hash(act))

    # make edges addition always in the same order
    dfg_edges = sorted(list(dfg.keys()))

    # represent edges
    for edge in dfg_edges:
        label = str(dfg[edge])
        viz.edge(str(hash(edge[0])), str(hash(edge[1])), fontsize=font_size, penwidth="2.0")

    start_activities_to_include = [act for act in start_activities if act in activities_map]
    end_activities_to_include = [act for act in end_activities if act in activities_map]

    if start_activities_to_include:
        viz.node("@@startnode", "START", style="filled", fillcolor="lightgrey", fontsize=font_size)
        for act in start_activities_to_include:
            label = str(start_activities[act]) if isinstance(start_activities, dict) and measure == "frequency" else ""
            viz.edge("@@startnode", activities_map[act], fontsize=font_size, penwidth="2.0")

    if end_activities_to_include:
        # <&#9632;>
        viz.node("@@endnode", "END", style="filled", fillcolor="lightgrey", fontsize=font_size)
        for act in end_activities_to_include:
            label = str(end_activities[act]) if isinstance(end_activities, dict) and measure == "frequency" else ""
            viz.edge(activities_map[act], "@@endnode", fontsize=font_size, penwidth="2.0")

    viz.attr(overlap='false')
    viz.attr(fontsize='11')

    viz.format = image_format.replace("html", "plain-ext")

    return viz


from pm4py.algo.discovery.inductive.fall_through.activity_once_per_trace import ActivityOncePerTraceUVCL
from powl.discovery.total_order_based.inductive.fall_through.activity_concurrent import POWLActivityConcurrentUVCL


class POWLActivityOncePerTraceUVCL(ActivityOncePerTraceUVCL, POWLActivityConcurrentUVCL):
    pass



from multiprocessing import Pool, Manager
from typing import Optional, Tuple, List, Dict, Any

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.algo.discovery.inductive.fall_through.flower import FlowerModelUVCL
from powl.objects.obj import OperatorPOWL
from pm4py.objects.process_tree.obj import Operator
from pm4py.util.compression import util as comut
from pm4py.util.compression.dtypes import UVCL


class POWLFlowerModelUVCL(FlowerModelUVCL):

    @classmethod
    def apply(cls, obj: IMDataStructureUVCL, pool: Pool = None, manager: Manager = None, parameters: Optional[Dict[str, Any]] = None) -> Optional[
        Tuple[OperatorPOWL, List[IMDataStructureUVCL]]]:
        log = obj.data_structure
        uvcl_do = UVCL()
        for a in comut.get_alphabet(log):
            uvcl_do[(a,)] = 1
        uvcl_redo = UVCL()
        im_uvcl_do = IMDataStructureUVCL(uvcl_do)
        im_uvcl_redo = IMDataStructureUVCL(uvcl_redo)
        return OperatorPOWL(Operator.LOOP, []), [im_uvcl_do, im_uvcl_redo]





import time
import uuid
from itertools import product

from powl.conversion.utils.pn_reduction import merge_places_connected_with_silent_transition, \
    merge_places_with_identical_preset_or_postset
from powl.objects.obj import Transition, SilentTransition, StrictPartialOrder, OperatorPOWL, FrequentTransition, \
    DecisionGraph

import pm4py.objects.conversion.process_tree.variants.to_petri_net as pt_to_pn
from pm4py.objects.petri_net.obj import Marking
from pm4py.objects.petri_net.obj import PetriNet
from pm4py.objects.petri_net.utils.petri_utils import add_arc_from_to, remove_place

from pm4py.objects.petri_net.utils import reduction
from pm4py.objects.process_tree.obj import Operator

REDUCE_IMPLICIT_PLACES = True


def merge_groups(a, b, groups):
    group_a = None
    group_b = None
    for group in groups:
        if a in group:
            group_a = group
        if b in group:
            group_b = group
    groups = [group for group in groups if group != group_a and group != group_b]
    groups.append(group_a.union(group_b))
    return groups


def add_transition_and_arcs(net, transition, p1, p2):
    net.transitions.add(transition)
    add_arc_from_to(p1, transition, net)
    add_arc_from_to(transition, p2, net)


def recursively_add_tree(powl, net, initial_entity_subtree, final_entity_subtree, counts, rec_depth,
                         flatten_frequency_tags, force_add_skip=False):

    if type(initial_entity_subtree) is PetriNet.Transition:
        initial_place = get_new_place(counts)
        net.places.add(initial_place)
        add_arc_from_to(initial_entity_subtree, initial_place, net)
    else:
        initial_place = initial_entity_subtree
    if final_entity_subtree is not None and type(final_entity_subtree) is PetriNet.Place:
        final_place = final_entity_subtree
    else:
        final_place = get_new_place(counts)
        net.places.add(final_place)
        if final_entity_subtree is not None and type(final_entity_subtree) is PetriNet.Transition:
            add_arc_from_to(final_place, final_entity_subtree, net)

    if force_add_skip:
        invisible = get_new_hidden_trans(counts, type_trans="skip")
        add_arc_from_to(initial_place, invisible, net)
        add_arc_from_to(invisible, final_place, net)

    if isinstance(powl, Transition):
        if isinstance(powl, SilentTransition):
            petri_trans = get_new_hidden_trans(counts, type_trans="skip")
            add_transition_and_arcs(net, petri_trans, initial_place, final_place)
        elif isinstance(powl, FrequentTransition):
            if flatten_frequency_tags:
                petri_trans = get_transition(counts, powl.activity, powl.activity)
                if powl.skippable:
                    silent_t = get_new_hidden_trans(counts, type_trans="skip")
                    if powl.selfloop:
                        add_transition_and_arcs(net, petri_trans, final_place, initial_place)
                        add_transition_and_arcs(net, silent_t, initial_place, final_place)
                    else:
                        add_transition_and_arcs(net, petri_trans, initial_place, final_place)
                        add_transition_and_arcs(net, silent_t, initial_place, final_place)
                else:
                    if powl.selfloop:
                        silent_t = get_new_hidden_trans(counts, type_trans="skip")
                        add_transition_and_arcs(net, petri_trans, initial_place, final_place)
                        add_transition_and_arcs(net, silent_t, final_place, initial_place)
                    else:
                        add_transition_and_arcs(net, petri_trans, initial_place, final_place)

            else:
                petri_trans = get_transition(counts, powl.label, powl.activity, powl.skippable, powl.selfloop)
                add_transition_and_arcs(net, petri_trans, initial_place, final_place)
        else:
            petri_trans = get_transition(counts, powl.label, powl.label)
            add_transition_and_arcs(net, petri_trans, initial_place, final_place)


    elif isinstance(powl, OperatorPOWL):
        tree_children = powl.children
        if powl.operator == Operator.XOR:
            for subtree in tree_children:
                net, counts, intermediate_place = recursively_add_tree(subtree, net, initial_place,
                                                                       final_place,
                                                                       counts,
                                                                       rec_depth + 1,
                                                                       flatten_frequency_tags)
        elif powl.operator == Operator.LOOP:
            new_initial_place = get_new_place(counts)
            net.places.add(new_initial_place)
            init_loop_trans = get_new_hidden_trans(counts, type_trans="init_loop")
            net.transitions.add(init_loop_trans)
            add_arc_from_to(initial_place, init_loop_trans, net)
            add_arc_from_to(init_loop_trans, new_initial_place, net)
            initial_place = new_initial_place
            loop_trans = get_new_hidden_trans(counts, type_trans="loop")
            net.transitions.add(loop_trans)

            exit_node = SilentTransition()
            do = tree_children[0]
            redo = tree_children[1]

            net, counts, int1 = recursively_add_tree(do, net, initial_place,
                                                     None, counts,
                                                     rec_depth + 1,
                                                     flatten_frequency_tags)
            net, counts, int2 = recursively_add_tree(redo, net, int1,
                                                     None, counts,
                                                     rec_depth + 1,
                                                     flatten_frequency_tags)
            net, counts, int3 = recursively_add_tree(exit_node, net, int1,
                                                     final_place, counts,
                                                     rec_depth + 1,
                                                     flatten_frequency_tags)

            looping_place = int2

            add_arc_from_to(looping_place, loop_trans, net)
            add_arc_from_to(loop_trans, initial_place, net)

    elif isinstance(powl, StrictPartialOrder):
        transitive_reduction = powl.order.get_transitive_reduction()
        tree_children = list(powl.children)
        tau_split = get_new_hidden_trans(counts, type_trans="tauSplit")
        net.transitions.add(tau_split)
        add_arc_from_to(initial_place, tau_split, net)
        tau_join = get_new_hidden_trans(counts, type_trans="tauJoin")
        net.transitions.add(tau_join)
        add_arc_from_to(tau_join, final_place, net)

        init_trans = []
        final_trans = []
        start_nodes = transitive_reduction.get_start_nodes()
        end_nodes = transitive_reduction.get_end_nodes()
        for subtree in tree_children:
            i_trans = get_new_hidden_trans(counts, type_trans="init_par")
            net.transitions.add(i_trans)
            if subtree in start_nodes:
                i_place = get_new_place(counts)
                net.places.add(i_place)
                add_arc_from_to(tau_split, i_place, net)

                add_arc_from_to(i_place, i_trans, net)

            f_trans = get_new_hidden_trans(counts, type_trans="final_par")
            net.transitions.add(f_trans)
            if subtree in end_nodes:
                f_place = get_new_place(counts)
                net.places.add(f_place)
                add_arc_from_to(f_trans, f_place, net)
                add_arc_from_to(f_place, tau_join, net)

            net, counts, intermediate_place = recursively_add_tree(subtree,
                                                                   net,
                                                                   i_trans,
                                                                   f_trans,
                                                                   counts,
                                                                   rec_depth + 1,
                                                                   flatten_frequency_tags)
            init_trans.append(i_trans)
            final_trans.append(f_trans)

        n = range(len(tree_children))
        for i, j in product(n, n):
            if transitive_reduction.is_edge_id(i, j):
                new_place = get_new_place(counts)
                net.places.add(new_place)
                add_arc_from_to(final_trans[i], new_place, net)
                add_arc_from_to(new_place, init_trans[j], net)

    elif isinstance(powl, DecisionGraph):

        real_children = powl.children # without artificial start/end nodes
        all_children = powl.order.nodes

        # create one unique input and one unique output places for each node
        node_to_pre_place = {}
        node_to_post_place = {}
        for child in real_children:
            pre_place = get_new_place(counts)
            post_place = get_new_place(counts)
            net.places.add(pre_place)
            net.places.add(post_place)
            node_to_pre_place[child] = pre_place
            node_to_post_place[child] = post_place
            net, counts, _ = recursively_add_tree(child, net, pre_place,
                                                  post_place, counts, rec_depth + 1,
                                                  flatten_frequency_tags)

        node_to_pre_place[powl.end] = final_place
        node_to_post_place[powl.start] = initial_place

        # Now create a silent transition for each edge
        for s in all_children:
            for t in all_children:
                if powl.order.is_edge(s, t):
                    silent = get_new_hidden_trans(counts)
                    net.transitions.add(silent)
                    add_arc_from_to(node_to_post_place[s], silent, net)
                    add_arc_from_to(silent, node_to_pre_place[t], net)

        return net, counts, final_place

    else:
        raise Exception("Unknown POWL operator!")

    return net, counts, final_place


def apply(powl, parameters=None):
    if parameters and "flatten_frequency_tags" in parameters:
        flatten_frequency_tags = parameters["flatten_frequency_tags"]
    else:
        flatten_frequency_tags = True

    counts = pt_to_pn.Counts()
    net = PetriNet('imdf_net_' + str(time.time()))
    initial_marking = Marking()
    final_marking = Marking()
    source = get_new_place(counts)
    source.name = "source"
    sink = get_new_place(counts)
    sink.name = "sink"
    net.places.add(source)
    net.places.add(sink)
    initial_marking[source] = 1
    final_marking[sink] = 1
    initial_mandatory = True
    final_mandatory = True
    if initial_mandatory:
        initial_place = get_new_place(counts)
        net.places.add(initial_place)
        tau_initial = get_new_hidden_trans(counts, type_trans="tau")
        net.transitions.add(tau_initial)
        add_arc_from_to(source, tau_initial, net)
        add_arc_from_to(tau_initial, initial_place, net)
    else:
        initial_place = source
    if final_mandatory:
        final_place = get_new_place(counts)
        net.places.add(final_place)
        tau_final = get_new_hidden_trans(counts, type_trans="tau")
        net.transitions.add(tau_final)
        add_arc_from_to(final_place, tau_final, net)
        add_arc_from_to(tau_final, sink, net)
    else:
        final_place = sink

    net, counts, last_added_place = recursively_add_tree(powl, net, initial_place, final_place, counts, 0,
                                                         flatten_frequency_tags=flatten_frequency_tags)

    if REDUCE_IMPLICIT_PLACES:

        net = merge_places_with_identical_preset_or_postset(net, source, sink)

        reduction.apply_simple_reduction(net)

        places = list(net.places)
        for place in places:
            if len(place.out_arcs) == 0 and place not in final_marking:
                remove_place(net, place)
            if len(place.in_arcs) == 0 and place not in initial_marking:
                remove_place(net, place)

        net = merge_places_connected_with_silent_transition(net, source, sink)

    return net, initial_marking, final_marking


def get_new_place(counts):
    """
    Create a new place in the Petri net
    """
    counts.inc_places()
    return PetriNet.Place('p_' + str(counts.num_places))


def get_new_hidden_trans(counts, type_trans="unknown"):
    """
    Create a new hidden transition in the Petri net
    """
    counts.inc_no_hidden()
    return PetriNet.Transition(type_trans + '_' + str(counts.num_hidden), None)


def get_transition(counts, label, activity, skippable=False, selfloop=False):
    """
    Create a transitions with the specified label in the Petri net
    """
    counts.inc_no_visible()
    return PetriNet.Transition(str(uuid.uuid4()), label, properties={"activity": activity,
                                                                     "skippable": skippable,
                                                                     "selfloop": selfloop})



from copy import copy

from powl.objects.BinaryRelation import BinaryRelation
from pm4py.objects.process_tree.obj import ProcessTree, Operator
from typing import List as TList, Optional, Union


class POWL(ProcessTree):
    def simplify_using_frequent_transitions(self) -> "POWL":
        return self

    def simplify(self) -> "POWL":
        return self


class Transition(POWL):
    transition_id: int = 0

    def __init__(self, label: Optional[str] = None) -> None:
        super().__init__()
        self._label = label
        self._identifier = Transition.transition_id
        Transition.transition_id = Transition.transition_id + 1

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Transition):
            return self._label == other._label and self._identifier == other._identifier
        return False

    def equal_content(self, other: object) -> bool:
        if isinstance(other, Transition):
            return self._label == other._label
        return False

    def __hash__(self) -> int:
        return self._identifier

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Transition):
            if self.label and other.label and self.label < other.label:
                return self.label < other.label
            return self._identifier < other._identifier
        elif isinstance(other, OperatorPOWL):
            return True
        elif isinstance(other, StrictPartialOrder):
            return True
        return NotImplemented


class SilentTransition(Transition):
    def __init__(self) -> None:
        super().__init__(label=None)


class FrequentTransition(Transition):
    def __init__(self, label, min_freq: Union[str, int], max_freq: Union[str, int]) -> None:
        self.skippable = False
        self.selfloop = False
        if min_freq == 0:
            self.skippable = True
        if max_freq == "-":
            self.selfloop = True
        self.activity = label
        if self.skippable or self.selfloop:
            label = str(label) + "\n" + "[" + str(min_freq) + "," + str(max_freq) + "]"

        super().__init__(label=label)

    def set_skippable(self, skip):
        self.skippable = skip
        self.__update_label()

    def set_selfloop(self, loop):
        self.selfloop = loop
        self.__update_label()

    def __update_label(self):
        if self.skippable:
            min_freq = 0
        else:
            min_freq = 1
        if self.selfloop:
            max_freq = "-"
        else:
            max_freq = 1
        self.label = str(self.activity) + "\n" + "[" + str(min_freq) + "," + str(max_freq) + "]"


class StrictPartialOrder(POWL):

    def __init__(self, nodes: TList[POWL]) -> None:
        super().__init__()
        self.operator = None
        self._set_order(nodes)
        self.additional_information = None

    def _set_order(self, nodes: TList[POWL]) -> None:
        self.order = BinaryRelation(nodes)

    def get_order(self) -> BinaryRelation:
        return self.order

    def _set_children(self, children: TList[POWL]) -> None:
        self.order.nodes = children

    def get_children(self) -> TList[POWL]:
        return self.order.nodes


    def __lt__(self, other: object) -> bool:
        if isinstance(other, StrictPartialOrder):
            return self.__repr__() < other.__repr__()
        elif isinstance(other, OperatorPOWL):
            return False
        elif isinstance(other, Transition):
            return False
        return NotImplemented

    partial_order = property(get_order, _set_order)
    children = property(get_children, _set_children)


    def equal_content(self, other: object) -> bool:
        if not isinstance(other, StrictPartialOrder):
            return False

        ordered_nodes_1 = sorted(list(self.order.nodes))
        ordered_nodes_2 = sorted(list(other.order.nodes))
        if len(ordered_nodes_1) != len(ordered_nodes_2):
            return False
        for i in range(len(ordered_nodes_1)):
            source_1 = ordered_nodes_1[i]
            source_2 = ordered_nodes_2[i]
            if not source_1.equal_content(source_2):
                return False
            for j in range(len(ordered_nodes_1)):
                target_1 = ordered_nodes_1[j]
                target_2 = ordered_nodes_2[j]
                if self.order.is_edge(source_1, target_1) and not other.order.is_edge(source_2, target_2):
                    return False
                if not self.order.is_edge(source_1, target_1) and other.order.is_edge(source_2, target_2):
                    return False
        return True

    def simplify_using_frequent_transitions(self) -> "StrictPartialOrder":
        new_nodes = {node: node.simplify_using_frequent_transitions() for node in self.children}
        res = StrictPartialOrder(list(new_nodes.values()))
        for node_1 in self.children:
            for node_2 in self.children:
                if self.partial_order.is_edge(node_1, node_2):
                    res.partial_order.add_edge(new_nodes[node_1], new_nodes[node_2])

        return res

    def simplify(self) -> "StrictPartialOrder":
        simplified_nodes = {}
        sub_nodes = {}
        start_nodes = {}
        end_nodes = {}

        def connected(node):
            for node2 in self.children:
                if self.partial_order.is_edge(node, node2) or self.partial_order.is_edge(node2, node):
                    return True
            return False

        for node_1 in self.children:
            simplified_node = node_1.simplify()
            if isinstance(simplified_node, StrictPartialOrder):

                if not connected(node_1):
                    sub_nodes[node_1] = simplified_node
                else:
                    s_nodes = simplified_node.order.get_start_nodes()
                    e_nodes = simplified_node.order.get_end_nodes()
                    if len(s_nodes) == 1 and len(e_nodes) == 1:
                        sub_nodes[node_1] = simplified_node
                        start_nodes[node_1] = list(s_nodes)[0]
                        end_nodes[node_1] = list(e_nodes)[0]
                    else:
                        simplified_nodes[node_1] = simplified_node
            else:
                simplified_nodes[node_1] = simplified_node

        new_nodes = list(simplified_nodes.values())
        for po, simplified_po in sub_nodes.items():
            new_nodes = new_nodes + list(simplified_po.children)
        res = StrictPartialOrder(new_nodes)
        for node_1 in self.children:
            for node_2 in self.children:
                if self.partial_order.is_edge(node_1, node_2):
                    if node_1 in simplified_nodes.keys() and node_2 in simplified_nodes.keys():
                        res.partial_order.add_edge(simplified_nodes[node_1], simplified_nodes[node_2])
                    elif node_1 in simplified_nodes.keys():
                        res.partial_order.add_edge(simplified_nodes[node_1], start_nodes[node_2])
                    elif node_2 in simplified_nodes.keys():
                        res.partial_order.add_edge(end_nodes[node_1], simplified_nodes[node_2])
                    else:
                        res.partial_order.add_edge(end_nodes[node_1], start_nodes[node_2])
        for po, simplified_po in sub_nodes.items():
            for node_1 in simplified_po.children:
                for node_2 in simplified_po.children:
                    if simplified_po.partial_order.is_edge(node_1, node_2):
                        res.partial_order.add_edge(node_1, node_2)
        return res


class Sequence(StrictPartialOrder):

    def __init__(self, nodes: TList[POWL]) -> None:
        super().__init__(nodes)
        for i in range(len(nodes)):
            for j in range(i + 1, len(nodes)):
                self.partial_order.add_edge(nodes[i], nodes[j])

    def simplify_using_frequent_transitions(self) -> "StrictPartialOrder":
        new_children = []
        for node in self.children:
            s_node = node.simplify_using_frequent_transitions()
            if isinstance(s_node, Sequence):
                new_children = new_children + s_node.children
            else:
                new_children = new_children + [s_node]
        res = Sequence(new_children)
        return res


class OperatorPOWL(POWL):
    def __init__(self, operator: Operator, children: TList[POWL]) -> None:
        super().__init__()
        self.operator = operator
        self.children = children

    def __lt__(self, other: object) -> bool:
        if isinstance(other, OperatorPOWL):
            return self.__repr__() < other.__repr__()
        elif isinstance(other, Transition):
            return False
        elif isinstance(other, StrictPartialOrder):
            return True
        return NotImplemented

    def equal_content(self, other: object) -> bool:
        if not isinstance(other, OperatorPOWL):
            return False

        if self.operator != other.operator:
            return False

        ordered_nodes_1 = sorted(list(self.children))
        ordered_nodes_2 = sorted(list(other.children))
        if len(ordered_nodes_1) != len(ordered_nodes_2):
            return False
        for i in range(len(ordered_nodes_1)):
            node_1 = ordered_nodes_1[i]
            node_2 = ordered_nodes_2[i]
            if not node_1.equal_content(node_2):
                return False
        return True

    def simplify_using_frequent_transitions(self) -> POWL:
        if self.operator is Operator.XOR and len(self.children) == 2:
            child_0 = self.children[0].simplify_using_frequent_transitions()
            child_1 = self.children[1].simplify_using_frequent_transitions()

            if isinstance(child_0, SilentTransition) and isinstance(child_1, SilentTransition):
                return child_0

            if isinstance(child_0, SilentTransition):
                if isinstance(child_1, FrequentTransition):
                    child_1.set_skippable(True)
                    return child_1
                elif isinstance(child_1, Transition):
                    return FrequentTransition(label=child_1.label, min_freq=0, max_freq=1)
            if isinstance(child_1, SilentTransition):
                if isinstance(child_0, FrequentTransition):
                    child_0.set_skippable(True)
                    return child_0
                elif isinstance(child_0, Transition):
                    return FrequentTransition(label=child_0.label, min_freq=0, max_freq=1)

        if self.operator is Operator.LOOP and len(self.children) == 2:
            child_0 = self.children[0]
            child_1 = self.children[1]
            if isinstance(child_0, Transition) and isinstance(child_1, SilentTransition):
                return FrequentTransition(label=child_0.label, min_freq=1, max_freq="-")
            elif isinstance(child_1, Transition) and isinstance(child_0, SilentTransition):
                return FrequentTransition(label=child_1.label, min_freq=0, max_freq="-")

        return OperatorPOWL(self.operator, [child.simplify_using_frequent_transitions() for child in self.children])

    def simplify(self) -> "OperatorPOWL":
        if self.operator is Operator.XOR and len(self.children) == 2:
            child_0 = self.children[0]
            child_1 = self.children[1]

            def merge_with_children(child0, child1):
                if isinstance(child0, SilentTransition) and isinstance(child1, OperatorPOWL) \
                        and child1.operator is Operator.LOOP:
                    if isinstance(child1.children[0], SilentTransition):
                        return OperatorPOWL(Operator.LOOP, [n.simplify() for n in child1.children])
                    elif isinstance(child1.children[1], SilentTransition):
                        return OperatorPOWL(Operator.LOOP, list(reversed([n.simplify() for n in child1.children])))

                return None

            res = merge_with_children(child_0, child_1)
            if res is not None:
                return res

            res = merge_with_children(child_1, child_0)
            if res is not None:
                return res

        if self.operator is Operator.XOR:
            new_children = []
            for child in self.children:
                s_child = child.simplify()
                if isinstance(s_child, OperatorPOWL) and s_child.operator is Operator.XOR:
                    for node in s_child.children:
                        new_children.append(node.simplify())
                else:
                    new_children.append(s_child)
            return OperatorPOWL(Operator.XOR, [child for child in new_children])
        else:
            return OperatorPOWL(self.operator, [child.simplify() for child in self.children])


class DecisionGraph(POWL):
    """
    A DecisionGraph is a POWL model defined over a set of nodes (each node is a POWL model)
    together with a binary relation (order) over these nodes, augmented with two artificial
    nodes: a start node and an end node.

    In the decision graph, each node represents a group (or branch) of activities (or submodels)
    and the binary relation encodes the allowed ordering between these nodes.
    """

    def __init__(self, order: BinaryRelation, start_nodes, end_nodes, empty_path=False) -> None:
        super().__init__()
        self.operator = None
        self.children = [n for n in order.nodes]
        self.start_nodes = start_nodes
        self.end_nodes = end_nodes
        # if not all(isinstance(node, POWL) for node in nodes):
        #     raise Exception("The nodes of the decision graph must be POWL models!")
        if not start_nodes or not set(start_nodes).issubset(order.nodes):
            raise Exception("Start nodes must be a non-empty subset of the nodes of the relation!")
        if not end_nodes or not set(end_nodes).issubset(order.nodes):
            raise Exception("End nodes must be a non-empty subset of the nodes of the relation!")
        self.start = StartNode()
        self.end = EndNode()
        order.add_node(self.start)
        order.add_node(self.end)
        for node in start_nodes:
            order.add_edge(self.start, node)
        for node in end_nodes:
            order.add_edge(node, self.end)
        if empty_path:
            order.add_edge(self.start, self.end)

        self.order = order

    def simplify(self) -> POWL:
        if len(self.children) == 1:
            child_0 = self.children[0]
            if self.order.is_edge(self.start, self.end):
                if isinstance(child_0, OperatorPOWL) and child_0.operator is Operator.LOOP:
                    if isinstance(child_0.children[0], SilentTransition):
                        return OperatorPOWL(Operator.LOOP, child_0.children).simplify()
                    elif isinstance(child_0.children[1], SilentTransition):
                        return OperatorPOWL(Operator.LOOP, list(reversed(child_0.children))).simplify()
                elif isinstance(child_0, DecisionGraph):
                    child_0.empty_path = True
                    child_0.order.add_edge(child_0.start, child_0.end)
                    return child_0.simplify()
            else:
                return child_0.simplify()

            new_dg = self

        else:
            new_dg = self.__group_pure_xor()

            seq = new_dg.__group_start_seq()
            if seq:
                return seq.simplify()

            seq = new_dg.__group_end_seq()
            if seq:
                return seq.simplify()

            res = new_dg.__group_pure_seq()
            if len(res.children) < len(new_dg.children):
                return res.simplify()

        new_children_map = {}
        for child in new_dg.children:
            s_child = child.simplify()
            new_children_map[child] = s_child
        return new_dg.__apply_mapping(new_children_map)

    def simplify_using_frequent_transitions(self) -> POWL:
        if len(self.children) == 1:
            child_0 = self.children[0]
            if isinstance(child_0, Transition):
                if self.order.is_edge(self.start, self.end):
                    return FrequentTransition(label=child_0.label, min_freq=0, max_freq=1)
                else:
                    return child_0

        new_children_map = {}
        edges_to_remove = set()
        for child in self.children:
            s_child = child.simplify_using_frequent_transitions()
            if isinstance(s_child, Transition):
                preset = self.order.get_preset(child)
                postset = self.order.get_postset(child)
                if all(self.order.is_edge(pre, post) for pre in preset for post in postset):
                    for pre in preset:
                        for post in postset:
                            edges_to_remove.add((pre, post))
                    if child in self.start_nodes:
                        self.start_nodes = [x for x in self.start_nodes if x not in postset]
                    if child in self.end_nodes:
                        self.end_nodes = [x for x in self.end_nodes if x not in preset]
                    if isinstance(s_child, FrequentTransition):
                        s_child.set_skippable(True)
                    else:
                        s_child = FrequentTransition(label=child.label, min_freq=0, max_freq=1)
            new_children_map[child] = s_child
        new_dg = self.__apply_mapping(new_children_map, edges_to_remove)

        old_size = len(new_dg.children) + 1
        while len(new_dg.children) < old_size:
            new_dg = new_dg.__group_pure_xor()
            old_size = len(new_dg.children)
            new_dg = new_dg.__group_pure_seq()

        seq1 = new_dg.__group_start_seq()
        start_nodes = []
        if seq1:
            start_nodes = seq1.children[:-1]
            new_dg = seq1.children[-1]

        seq2 = new_dg.__group_end_seq()
        end_nodes = []
        if seq2:
            end_nodes = seq2.children[1:]
            new_dg = seq2.children[0]

        new_node = new_dg
        if len(new_dg.children) == 1:
            child_0 = new_dg.children[0]
            if new_dg.order.is_edge(new_dg.start, new_dg.end):
                if isinstance(child_0, OperatorPOWL) and child_0.operator is Operator.LOOP:
                    if isinstance(child_0.children[0], SilentTransition):
                        new_node = OperatorPOWL(Operator.LOOP, child_0.children)
                    elif isinstance(child_0.children[1], SilentTransition):
                        new_node = OperatorPOWL(Operator.LOOP, list(reversed(child_0.children)))
                elif isinstance(child_0, DecisionGraph):
                    child_0.empty_path = True
                    child_0.order.add_edge(child_0.start, child_0.end)
                    new_node = child_0
            else:
                new_node = child_0

        if len(start_nodes) > 0 or len(end_nodes) > 0:
            if isinstance(new_node, Sequence):
                return Sequence(start_nodes + new_node.children + end_nodes)
            else:
                return Sequence(start_nodes + [new_node] + end_nodes)
        else:
            return new_node

    def __apply_mapping(self, mapping, edges_to_remove=None) -> "DecisionGraph":
        if edges_to_remove is None:
            edges_to_remove = set()
        res = BinaryRelation(list(set(mapping.values())))
        for src in self.children:
            for tgt in self.children:
                if self.order.is_edge(src, tgt) and (src, tgt) not in edges_to_remove:
                    new_src = mapping[src]
                    new_tgt = mapping[tgt]
                    if new_src != new_tgt:
                        res.add_edge(new_src, new_tgt)
        new_start_nodes = list({mapping[child] for child in self.start_nodes})
        new_end_nodes = list({mapping[child] for child in self.end_nodes})
        empty_path = self.order.is_edge(self.start, self.end) and not (self.start, self.end) in edges_to_remove
        return DecisionGraph(res, new_start_nodes, new_end_nodes, empty_path)

    def __create_mapping(self, old_children, new_child):
        mapping = {}
        for key in self.children:
            if key in old_children:
                mapping[key] = new_child
            else:
                mapping[key] = key
        return mapping

    def __group_pure_xor(self):
        new_dg = copy(self)
        visited = set()
        for child in list(self.children):
            if child not in visited:
                visited.add(child)
                post1 = new_dg.order.get_postset(child)
                pre1 = new_dg.order.get_preset(child)
                if len(pre1) > 1 or len(post1) > 1:
                    candidates_for_xor = [child2 for child2 in new_dg.children if
                                          pre1 == new_dg.order.get_preset(child2) and post1 == new_dg.order.get_postset(
                                              child2)]
                    if len(candidates_for_xor) > 1:
                        visited.update(candidates_for_xor)

                        new_rel = BinaryRelation(candidates_for_xor)
                        empty_path = False
                        edges_to_remove = None
                        if all(new_dg.order.is_edge(pre, post) for pre in pre1 for post in post1):
                            empty_path = True
                            edges_to_remove = [(pre, post) for pre in pre1 for post in post1]
                        dg_child = DecisionGraph(new_rel, candidates_for_xor, candidates_for_xor, empty_path)
                        mapping = new_dg.__create_mapping(candidates_for_xor, dg_child)
                        new_dg = new_dg.__apply_mapping(mapping, edges_to_remove)
        return new_dg

    def __group_pure_seq(self):
        for child in list(self.children):
            for child2 in self.children:
                post1 = self.order.get_postset(child)
                pre2 = self.order.get_preset(child2)
                if pre2 == {child} and post1 == {child2}:
                    seq = Sequence([child, child2])
                    mapping = self.__create_mapping({child, child2}, seq)
                    new_dg = self.__apply_mapping(mapping)
                    return new_dg.__group_pure_seq()
        return self

    def __group_start_seq(self):
        start_list = []
        current_dg = self
        while len(current_dg.children) > 1 and len(current_dg.start_nodes) == 1 and not current_dg.order.is_edge(
                current_dg.start, current_dg.end):
            start = current_dg.start_nodes[0]
            start_list.append(start)
            postset = current_dg.order.get_postset(start)
            new_start_nodes = list(postset - {current_dg.end})
            new_children = [n for n in current_dg.children if n != start]
            new_end_nodes = [n for n in current_dg.end_nodes if n != start]
            new_order = BinaryRelation(new_children)
            for c1 in new_children:
                for c2 in new_children:
                    if current_dg.order.is_edge(c1, c2):
                        new_order.add_edge(c1, c2)
            empty_path = current_dg.end in postset
            current_dg = DecisionGraph(new_order, new_start_nodes, new_end_nodes, empty_path)
        if len(start_list) > 0:
            seq = Sequence(start_list + [current_dg])
            return seq
        return None

    def __group_end_seq(self):
        end_list = []
        current_dg = self
        while len(current_dg.children) > 1 and len(current_dg.end_nodes) == 1 and not current_dg.order.is_edge(
                current_dg.start, current_dg.end):
            end = current_dg.end_nodes[0]
            end_list = [end] + end_list
            pretset = current_dg.order.get_preset(end)
            new_end_nodes = list(pretset - {current_dg.start})
            new_children = [n for n in current_dg.children if n != end]
            new_start_nodes = [n for n in current_dg.start_nodes if n != end]
            new_order = BinaryRelation(new_children)
            for c1 in new_children:
                for c2 in new_children:
                    if current_dg.order.is_edge(c1, c2):
                        new_order.add_edge(c1, c2)
            empty_path = current_dg.start in pretset
            current_dg = DecisionGraph(new_order, new_start_nodes, new_end_nodes, empty_path)
        if len(end_list) > 0:
            seq = Sequence([current_dg] + end_list)
            return seq
        return None


class StartNode:
    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        return "⊕start"


class EndNode:
    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        return "⊕end"



from itertools import combinations
from typing import Optional, Tuple, List, TypeVar, Generic, Dict, Any, Type

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.algo.discovery.inductive.fall_through.empty_traces import EmptyTracesUVCL
from pm4py.algo.discovery.inductive.variants.imf import IMFParameters
from pm4py.algo.discovery.inductive.variants.abc import InductiveMinerFramework
from pm4py.util import exec_utils
from pm4py.objects.dfg.obj import DFG

from powl.objects.obj import POWL, StrictPartialOrder, DecisionGraph, OperatorPOWL
from powl.discovery.total_order_based.inductive.fall_through.empty_traces import POWLEmptyTracesUVCL
from powl.discovery.total_order_based.inductive.base_case.factory import BaseCaseFactory
from powl.discovery.total_order_based.inductive.cuts.factory import CutFactory
from powl.discovery.total_order_based.inductive.fall_through.factory import FallThroughFactory
from powl.discovery.total_order_based.inductive.utils.filtering import FILTERING_TYPE, FilteringType, \
    filter_most_frequent_variants, FILTERING_THRESHOLD, filter_most_frequent_variants_with_decreasing_factor
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import POWLDiscoveryVariant
from powl.objects.BinaryRelation import BinaryRelation
from pm4py.objects.process_tree.obj import Operator

from copy import copy


T = TypeVar('T', bound=IMDataStructureUVCL)


class IMBasePOWL(Generic[T], InductiveMinerFramework[T]):

    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.TREE

    def empty_traces_cut(self) -> Type[EmptyTracesUVCL]:
        return POWLEmptyTracesUVCL

    def apply(self, obj: IMDataStructureUVCL, parameters: Optional[Dict[str, Any]] = None, second_iteration: bool = False) -> POWL:

        # empty_traces = self.empty_traces_cut().apply(obj, parameters)
        # if empty_traces is not None:
        #     return self._recurse(empty_traces[0], empty_traces[1], parameters)

        noise_threshold = exec_utils.get_param_value(IMFParameters.NOISE_THRESHOLD, parameters, 0.0)

        empty_traces = self.empty_traces_cut().apply(obj, parameters)
        if empty_traces is not None:
            number_original_traces = sum(y for y in obj.data_structure.values())
            number_filtered_traces = sum(y for y in empty_traces[1][-1].data_structure.values())

            if number_original_traces - number_filtered_traces > noise_threshold * number_original_traces:
                return self._recurse(empty_traces[0], empty_traces[1], parameters)
            else:
                obj = empty_traces[1][-1]

        powl = self.apply_base_cases(obj, parameters)
        if powl is not None:
            return powl

        cut = self.find_cut(obj, parameters)
        if cut is not None:
            powl = self._recurse(cut[0], cut[1], parameters=parameters)

        if powl is not None:
            return powl

        if FILTERING_TYPE in parameters.keys():
            filtering_type = parameters[FILTERING_TYPE]
            if filtering_type not in FilteringType:
                raise KeyError("Invalid FILTERING_TYPE: " + str(filtering_type))

            if filtering_type is FilteringType.DFG_FREQUENCY:
                if not second_iteration:
                    noise_threshold = exec_utils.get_param_value(IMFParameters.NOISE_THRESHOLD, parameters, 0.0)
                    filtered_ds = self.__filter_dfg_noise(obj, noise_threshold)
                    tree = self.apply(filtered_ds, parameters=parameters, second_iteration=True)
                    if tree is not None:
                        return tree

            elif filtering_type is FilteringType.DYNAMIC:
                filtered_log = filter_most_frequent_variants(obj.data_structure)
                if len(filtered_log.data_structure) > 0:
                    return self.apply(filtered_log, parameters=parameters)

            elif filtering_type is FilteringType.DECREASING_FACTOR:
                if FILTERING_THRESHOLD in parameters.keys():
                    t = parameters[FILTERING_THRESHOLD]
                    if isinstance(t, float) and 0 <= t < 1:
                        t = [t]
                    if isinstance(t, list):
                        for factor in t:
                            if factor > 0:
                                    filtered_log = filter_most_frequent_variants_with_decreasing_factor(obj.data_structure,                                                                                                decreasing_factor=factor)
                                    if len(filtered_log.data_structure) == 0:
                                        break
                                    elif len(filtered_log.data_structure) < len(obj.data_structure):
                                        return self.apply(filtered_log, parameters=parameters)
                    else:
                        raise KeyError("Invalid filtering threshold!")
            else:
                raise KeyError("Invalid filtering type!")

        ft = self.fall_through(obj, parameters)
        return self._recurse(ft[0], ft[1], parameters=parameters)

    def apply_base_cases(self, obj: T, parameters: Optional[Dict[str, Any]] = None) -> Optional[POWL]:
        return BaseCaseFactory.apply_base_cases(obj, parameters=parameters)

    def find_cut(self, obj: T, parameters: Optional[Dict[str, Any]] = None) -> Optional[Tuple[POWL, List[T]]]:
        return CutFactory.find_cut(obj, parameters=parameters)

    def fall_through(self, obj: T, parameters: Optional[Dict[str, Any]] = None) -> Tuple[POWL, List[T]]:
        return FallThroughFactory.fall_through(obj, self._pool, self._manager, parameters=parameters)

    def _recurse(self, powl: POWL, objs: List[T], parameters: Optional[Dict[str, Any]] = None):
        children = [self.apply(obj, parameters=parameters) for obj in objs]
        if isinstance(powl, StrictPartialOrder):
            powl_new = StrictPartialOrder(children)
            for i, j in combinations(range(len(powl.children)), 2):
                if powl.order.is_edge_id(i, j):
                    powl_new.order.add_edge(children[i], children[j])
                elif powl.order.is_edge_id(j, i):
                    powl_new.order.add_edge(children[j], children[i])
            return powl_new
        elif isinstance(powl, DecisionGraph):
            new_order = BinaryRelation(children)
            for i, j in combinations(range(len(powl.children)), 2):
                if powl.order.is_edge(objs[i], objs[j]):
                    new_order.add_edge(children[i], children[j])
                elif powl.order.is_edge(objs[j], objs[i]):
                    new_order.add_edge(children[j], children[i])
            start_nodes = [children[i] for i in range(len(powl.children)) if objs[i] in powl.start_nodes]
            end_nodes = [children[i] for i in range(len(powl.children)) if objs[i] in powl.end_nodes]
            empty_path = powl.order.is_edge(powl.start, powl.end)
            return DecisionGraph(new_order, start_nodes, end_nodes, empty_path=empty_path)
        elif isinstance(powl, OperatorPOWL):
            if powl.operator == Operator.LOOP and len(children) > 2:
                new_child = OperatorPOWL(Operator.XOR, children[1:])
                children = [children[0], new_child]
            powl.children.extend(children)
            return powl
        else:
            raise Exception("Unsupported POWL type!")

    def __filter_dfg_noise(self, obj, noise_threshold):
        start_activities = copy(obj.dfg.start_activities)
        end_activities = copy(obj.dfg.end_activities)
        dfg = copy(obj.dfg.graph)
        outgoing_max_occ = {}
        for x, y in dfg.items():
            act = x[0]
            if act not in outgoing_max_occ:
                outgoing_max_occ[act] = y
            else:
                outgoing_max_occ[act] = max(y, outgoing_max_occ[act])
            if act in end_activities:
                outgoing_max_occ[act] = max(outgoing_max_occ[act], end_activities[act])
        dfg_list = sorted([(x, y) for x, y in dfg.items()], key=lambda x: (x[1], x[0]), reverse=True)
        dfg_list = [x for x in dfg_list if x[1] > noise_threshold * outgoing_max_occ[x[0][0]]]
        dfg_list = [x[0] for x in dfg_list]
        # filter the elements in the DFG
        graph = {x: y for x, y in dfg.items() if x in dfg_list}

        dfg = DFG()
        for sa in start_activities:
            dfg.start_activities[sa] = start_activities[sa]
        for ea in end_activities:
            dfg.end_activities[ea] = end_activities[ea]
        for act in graph:
            dfg.graph[act] = graph[act]

        return IMDataStructureUVCL(obj.data_structure, dfg)



import networkx as nx

from powl.discovery.partial_order_based.utils.constants import VARIANT_FREQUENCY_KEY
from powl.discovery.partial_order_based.utils.simplified_objects import Graph, get_leaves
from pm4py.algo.discovery.inductive.cuts import utils as cut_util


class XORMiner:

    @classmethod
    def find_disjoint_activities(cls, partial_orders, all_activity_labels):
        """
        Finds activities that never occur together in the same partial order.

        :param all_activity_labels:
        :param partial_orders: List of partial orders.
        :return: Groups of disjoint activities.
        """

        all_activity_labels = sorted(all_activity_labels)
        adjacency = {activity: {other_activity: 0 for other_activity in all_activity_labels} for activity in all_activity_labels}

        for graph in partial_orders:
            activity_labels_in_trace = get_leaves(graph)
            for i, activity in enumerate(activity_labels_in_trace):
                for j, other_activity in enumerate(activity_labels_in_trace):
                    if i != j:
                        adjacency[activity][other_activity] += 1

        found_xor = False
        clusters = [[a] for a in all_activity_labels]
        for i in range(len(all_activity_labels)):
            activity = all_activity_labels[i]
            for j in range(i + 1, len(all_activity_labels)):
                other_activity = all_activity_labels[j]
                if adjacency[activity][other_activity] == 0 and adjacency[other_activity][activity] == 0:
                    found_xor = True
                    clusters = cut_util.merge_lists_based_on_activities(activity, other_activity, clusters)

        if found_xor:
            res = []
            for cluster in clusters:
                if len(cluster) == 1:
                    pass
                else:
                    from itertools import combinations
                    nx_graph = nx.DiGraph()
                    nx_graph.add_nodes_from(cluster)
                    for a, b in combinations(cluster, 2):
                        if adjacency[a][b] > 0 and adjacency[b][a] > 0:
                            nx_graph.add_edge(a, b)
                    nx_und = nx_graph.to_undirected()
                    conn_comps = [nx_und.subgraph(c).copy() for c in nx.connected_components(nx_und)]
                    if len(conn_comps) > 1:
                        cuts = list()
                        for comp in conn_comps:
                            cuts.append(set(comp.nodes))
                        res.append(cuts)
                    else:
                        return None
            return res
        else:
            return None

    @classmethod
    def project_partial_orders_on_groups(cls, partial_orders, group):
        res = []
        for graph in partial_orders:
            new_nodes = frozenset([n for n in graph.nodes if get_leaves(n).issubset(group)])
            if len(new_nodes) == 0:
                continue
            new_edges = frozenset([(s, t) for (s, t) in graph.edges if s in new_nodes and t in new_nodes])
            new_graph = Graph(new_nodes, new_edges)
            found = False
            if not found:
                new_graph.additional_information = {VARIANT_FREQUENCY_KEY: graph.additional_information[VARIANT_FREQUENCY_KEY]}
                res.append(new_graph)
        return res

    @classmethod
    def apply_mapping(cls, orders, label_mapping):
        res = []

        for graph in orders:
            new_nodes = set()
            for node in graph.nodes:
                label = get_leaves(node).pop()
                if label in label_mapping.keys():
                    new_nodes.add(label_mapping[node.label])
                else:
                    new_nodes.add(node)

            edges_set = set()
            for (s, t) in graph.edges:
                XORMiner.__add_edge(s, t, label_mapping, edges_set)

            for s in new_nodes:
                for t in new_nodes:
                    if (s, t) in edges_set and (t, s) in edges_set:
                        edges_set.remove((s, t))
                        edges_set.remove((t, s))

            new_graph = Graph(frozenset(new_nodes), frozenset(edges_set), graph.additional_information)

            found_po = False
            if not found_po:
                res.append(new_graph)
        return res

    @classmethod
    def __add_edge(cls, node, other_node, label_mapping, edges_set):
        label = get_leaves(node).pop()
        if label in label_mapping.keys():
            source = label_mapping[label]
        else:
            source = node
        other_labels = get_leaves(other_node)
        if other_labels.issubset(label_mapping.keys()):
            target = label_mapping[other_labels.pop()]
        else:
            target = other_node
        if source != target:
            edges_set.add((source, target))
